from mot.lib.cl_function import CLFunction, SimpleCLFunction
from mot.lib.utils import split_cl_function

__author__ = 'Robbert Harms'
__date__ = '2018-08-01'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


return_code_labels = {
    0: ['default', 'no return code specified'],
    1: ['found zero', 'sum of squares/evaluation below underflow limit'],
    2: ['converged', 'the relative error in the sum of squares/evaluation is at most tol'],
    3: ['converged', 'the relative error of the parameter vector is at most tol'],
    4: ['converged', 'both errors are at most tol'],
    5: ['trapped', 'by degeneracy; increasing epsilon might help'],
    6: ['exhausted', 'number of function calls exceeding preset patience'],
    7: ['failed', 'ftol<tol: cannot reduce sum of squares any further'],
    8: ['failed', 'xtol<tol: cannot improve approximate solution any further'],
    9: ['failed', 'gtol<tol: cannot improve approximate solution any further'],
    10: ['NaN', 'Function value is not-a-number or infinite'],
    11: ['exhausted', 'temperature decreased to 0.0']
}


class OptimizeResults(dict):
    """Represents the optimization results.

    Attributes:
        x (ndarray): the optimized parameter maps, an (d, p) array with for d problems a value for every p parameters
        status (ndarray): the return codes, an (d,) vector with for d problems the status return code
    """
    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __repr__(self):
        if self.keys():
            m = max(map(len, list(self.keys()))) + 1
            return '\n'.join([k.rjust(m) + ': ' + repr(v)
                              for k, v in sorted(self.items())])
        else:
            return self.__class__.__name__ + "()"

    def __dir__(self):
        return list(self.keys())


class ConstraintFunction(CLFunction):
    """These functions are meant to be provided to the optimization routines.

    If provided to the optimization routines, they should hold a CL function with the signature:

    .. code-block:: c

        void <func_name>(local const mot_float_type* const x,
                         void* data,
                         local mot_float_type* constraints);

    Although this is not enforced for general usage of this class.

    Since the number of constraints in the ``constraints`` array is variable, this class additionally specifies the
    number of constraints using the method :meth:`get_nmr_constraints`.
    """

    def get_nmr_constraints(self):
        """Get the number of constraints defined in this function.

        Returns:
            int: the number of constraints defined in this function.
        """
        raise NotImplementedError()


class SimpleConstraintFunction(SimpleCLFunction, ConstraintFunction):

    def __init__(self, *args, nmr_constraints=None, **kwargs):
        super().__init__(*args, **kwargs)
        self._nmr_constraints = nmr_constraints
        if nmr_constraints is None:
            raise ValueError('Number of constraints not defined.')

    @classmethod
    def from_string(cls, cl_function, dependencies=(), nmr_constraints=None):
        """Parse the given CL function into a SimpleCLFunction object.

        Args:
            cl_function (str): the function we wish to turn into an object
            dependencies (list or tuple of CLLibrary): The list of CL libraries this function depends on

        Returns:
            SimpleCLFunction: the CL data type for this parameter declaration
        """
        return_type, function_name, parameter_list, body = split_cl_function(cl_function)
        return SimpleConstraintFunction(return_type, function_name, parameter_list, body, dependencies=dependencies,
                                        nmr_constraints=nmr_constraints)

    def get_nmr_constraints(self):
        return self._nmr_constraints



from mot.lib.cl_function import SimpleCLFunction
from mot.configuration import CLRuntimeInfo
from mot.lib.kernel_data import Array, Scalar, CompositeArray, Struct, LocalMemory
from mot.lib.utils import all_elements_equal, get_single_value
from mot.library_functions import Powell, Subplex, NMSimplex, LevenbergMarquardt
from mot.optimize.base import OptimizeResults
import numpy as np

__author__ = 'Robbert Harms'
__date__ = '2018-08-01'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


def minimize(func, x0, data=None, method=None, lower_bounds=None, upper_bounds=None, constraints_func=None,
             nmr_observations=None, cl_runtime_info=None, options=None):
    """Minimization of one or more variables.

    For an easy wrapper of function maximization, see :func:`maximize`.

    All boundary conditions are enforced using the penalty method. That is, we optimize the objective function:

    .. math::

        F(x) = f(x) \mu \sum \max(0, g_i(x))^2

    where :math:`F(x)` is the new objective function, :math:`f(x)` is the old objective function, :math:`g_i` are
    the boundary functions defined as :math:`g_i(x) \leq 0` and :math:`\mu` is the penalty weight.

    The penalty weight is by default :math:`\mu = 1e20` and can be set
    using the ``options`` dictionary as ``penalty_weight``.

    Args:
        func (mot.lib.cl_function.CLFunction): A CL function with the signature:

            .. code-block:: c

                double <func_name>(local const mot_float_type* const x,
                                   void* data,
                                   local mot_float_type* objective_list);

            The objective list needs to be filled when the provided pointer is not null. It should contain
            the function values for each observation. This list is used by non-linear least-squares routines,
            and will be squared by the least-square optimizer. This is only used by the ``Levenberg-Marquardt`` routine.

        x0 (ndarray): Initial guess. Array of real elements of size (n, p), for 'n' problems and 'p'
            independent variables.
        data (mot.lib.kernel_data.KernelData): the kernel data we will load. This is returned to the likelihood function
            as the ``void* data`` pointer.
        method (str): Type of solver.  Should be one of:
            - 'Levenberg-Marquardt'
            - 'Nelder-Mead'
            - 'Powell'
            - 'Subplex'

            If not given, defaults to 'Powell'.
        lower_bounds (tuple): per parameter a lower bound, if given, the optimizer ensures ``a <= x`` with
            a the lower bound and x the parameter. If not given, -infinity is assumed for all parameters.
            Each tuple element can either be a scalar or a vector. If a vector is given the first dimension length
            should match that of the parameters.
        upper_bounds (tuple): per parameter an upper bound, if given, the optimizer ensures ``x >= b`` with
            b the upper bound and x the parameter. If not given, +infinity is assumed for all parameters.
            Each tuple element can either be a scalar or a vector. If a vector is given the first dimension length
            should match that of the parameters.
        constraints_func (mot.optimize.base.ConstraintFunction): function to compute (inequality) constraints.
            Should hold a CL function with the signature:

            .. code-block:: c

                void <func_name>(local const mot_float_type* const x,
                                 void* data,
                                 local mot_float_type* constraints);

            Where ``constraints_values`` is filled as:

            .. code-block:: c

                constraints[i] = g_i(x)

            That is, for each constraint function :math:`g_i`, formulated as :math:`g_i(x) <= 0`, we should return
            the function value of :math:`g_i`.

        nmr_observations (int): the number of observations returned by the optimization function.
            This is only needed for the ``Levenberg-Marquardt`` method.
        cl_runtime_info (mot.configuration.CLRuntimeInfo): the CL runtime information
        options (dict): A dictionary of solver options. All methods accept the following generic options:
            - patience (int): Maximum number of iterations to perform.
            - penalty_weight (float): the weight of the penalty term for the boundary conditions

    Returns:
        mot.optimize.base.OptimizeResults:
            The optimization result represented as a ``OptimizeResult`` object.
            Important attributes are: ``x`` the solution array.
    """
    if not method:
        method = 'Powell'

    cl_runtime_info = cl_runtime_info or CLRuntimeInfo()

    if len(x0.shape) < 2:
        x0 = x0[..., None]

    lower_bounds = _bounds_to_array(lower_bounds or np.ones(x0.shape[1]) * -np.inf)
    upper_bounds = _bounds_to_array(upper_bounds or np.ones(x0.shape[1]) * np.inf)

    if method == 'Powell':
        return _minimize_powell(func, x0, cl_runtime_info, lower_bounds, upper_bounds,
                                constraints_func=constraints_func, data=data, options=options)
    elif method == 'Nelder-Mead':
        return _minimize_nmsimplex(func, x0, cl_runtime_info, lower_bounds, upper_bounds,
                                   constraints_func=constraints_func, data=data, options=options)
    elif method == 'Levenberg-Marquardt':
        return _minimize_levenberg_marquardt(func, x0, nmr_observations, cl_runtime_info, lower_bounds, upper_bounds,
                                             constraints_func=constraints_func, data=data, options=options)
    elif method == 'Subplex':
        return _minimize_subplex(func, x0, cl_runtime_info, lower_bounds, upper_bounds,
                                 constraints_func=constraints_func, data=data, options=options)
    raise ValueError('Could not find the specified method "{}".'.format(method))


def _bounds_to_array(bounds):
    """Create a CompositeArray to hold the bounds."""
    elements = []
    for value in bounds:
        if all_elements_equal(value):
            elements.append(Scalar(get_single_value(value), ctype='mot_float_type'))
        else:
            elements.append(Array(value, ctype='mot_float_type', as_scalar=True))
    return CompositeArray(elements, 'mot_float_type', address_space='local')


def maximize(func, x0, nmr_observations, **kwargs):
    """Maximization of a function.

    This wraps the objective function to take the negative of the computed values and passes it then on to one
    of the minimization routines.

    Args:
        func (mot.lib.cl_function.CLFunction): A CL function with the signature:

            .. code-block:: c

                double <func_name>(local const mot_float_type* const x,
                                   void* data,
                                   local mot_float_type* objective_list);

            The objective list needs to be filled when the provided pointer is not null. It should contain
            the function values for each observation. This list is used by non-linear least-squares routines,
            and will be squared by the least-square optimizer. This is only used by the ``Levenberg-Marquardt`` routine.

        x0 (ndarray): Initial guess. Array of real elements of size (n, p), for 'n' problems and 'p'
            independent variables.
        nmr_observations (int): the number of observations returned by the optimization function.
        **kwargs: see :func:`minimize`.
    """
    wrapped_func = SimpleCLFunction.from_string('''
        double _negate_''' + func.get_cl_function_name() + '''(
                local mot_float_type* x,
                void* data, 
                local mot_float_type* objective_list){

            double return_val = ''' + func.get_cl_function_name() + '''(x, data, objective_list);    

            if(objective_list){
                const uint nmr_observations = ''' + str(nmr_observations) + ''';
                uint local_id = get_local_id(0);
                uint workgroup_size = get_local_size(0);

                uint observation_ind;
                for(uint i = 0; i < (nmr_observations + workgroup_size - 1) / workgroup_size; i++){
                    observation_ind = i * workgroup_size + local_id;

                    if(observation_ind < nmr_observations){
                        objective_list[observation_ind] *= -1;    
                    }
                }
            }
            return -return_val;
        }
    ''', dependencies=[func])
    kwargs['nmr_observations'] = nmr_observations
    return minimize(wrapped_func, x0, **kwargs)


def get_minimizer_options(method):
    """Return a dictionary with the default options for the given minimization method.

    Args:
        method (str): the name of the method we want the options off

    Returns:
        dict: a dictionary with the default options
    """
    if method == 'Powell':
        return {'patience': 2,
                'patience_line_search': None,
                'reset_method': 'EXTRAPOLATED_POINT'}

    elif method == 'Nelder-Mead':
        return {'patience': 200,
                'alpha': 1.0, 'beta': 0.5, 'gamma': 2.0, 'delta': 0.5, 'scale': 0.1,
                'adaptive_scales': True}

    elif method == 'Levenberg-Marquardt':
        return {'patience': 250, 'step_bound': 100.0, 'scale_diag': 1, 'usertol_mult': 30}

    elif method == 'Subplex':
        return {'patience': 10,
                'patience_nmsimplex': 100,
                'alpha': 1.0, 'beta': 0.5, 'gamma': 2.0, 'delta': 0.5, 'scale': 1.0, 'psi': 0.0001, 'omega': 0.01,
                'adaptive_scales': True,
                'min_subspace_length': 'auto',
                'max_subspace_length': 'auto'}

    raise ValueError('Could not find the specified method "{}".'.format(method))


def _clean_options(method, provided_options):
    """Clean the given input options.

    This will make sure that all options are present, either with their default values or with the given values,
    and that no other options are present then those supported.

    Args:
        method (str): the method name
        provided_options (dict): the given options

    Returns:
        dict: the resulting options dictionary
    """
    provided_options = provided_options or {}
    default_options = get_minimizer_options(method)

    result = {}

    for name, default in default_options.items():
        if name in provided_options:
            result[name] = provided_options[name]
        else:
            result[name] = default_options[name]
    return result


def _minimize_powell(func, x0, cl_runtime_info, lower_bounds, upper_bounds,
                     constraints_func=None, data=None, options=None):
    """
    Options:
        patience (int): Used to set the maximum number of iterations to patience*(number_of_parameters+1)
        reset_method (str): one of 'EXTRAPOLATED_POINT' or 'RESET_TO_IDENTITY' lower case or upper case.
        patience_line_search (int): the patience of the searching algorithm. Defaults to the
            same patience as for the Powell algorithm itself.
    """
    options = options or {}
    nmr_problems = x0.shape[0]
    nmr_parameters = x0.shape[1]

    penalty_data, penalty_func = _get_penalty_function(nmr_parameters, constraints_func)

    eval_func = SimpleCLFunction.from_string('''
        double evaluate(local mot_float_type* x, void* data){
            double penalty = _mle_penalty(
                x, 
                ((_powell_eval_func_data*)data)->data,
                ((_powell_eval_func_data*)data)->lower_bounds,
                ((_powell_eval_func_data*)data)->upper_bounds,
                ''' + str(options.get('penalty_weight', 1e30)) + ''',  
                ((_powell_eval_func_data*)data)->penalty_data
            );
            
            double func_val = ''' + func.get_cl_function_name() + '''(x, ((_powell_eval_func_data*)data)->data, 0);
            
            if(isnan(func_val)){
                return INFINITY;
            }
            
            return func_val + penalty;
        }
    ''', dependencies=[func, penalty_func])

    optimizer_func = Powell(eval_func, nmr_parameters, **_clean_options('Powell', options))

    kernel_data = {'model_parameters': Array(x0, ctype='mot_float_type', mode='rw'),
                   'data': Struct({'data': data,
                                   'lower_bounds': lower_bounds,
                                   'upper_bounds': upper_bounds,
                                   'penalty_data': penalty_data}, '_powell_eval_func_data')}
    kernel_data.update(optimizer_func.get_kernel_data())

    return_code = optimizer_func.evaluate(
        kernel_data, nmr_problems,
        use_local_reduction=all(env.is_gpu for env in cl_runtime_info.cl_environments),
        cl_runtime_info=cl_runtime_info)

    return OptimizeResults({'x': kernel_data['model_parameters'].get_data(),
                            'status': return_code})


def _minimize_nmsimplex(func, x0, cl_runtime_info, lower_bounds, upper_bounds,
                        constraints_func=None, data=None, options=None):
    """Use the Nelder-Mead simplex method to calculate the optimimum.

    The scales should satisfy the following constraints:

        .. code-block:: python

            alpha > 0
            0 < beta < 1
            gamma > 1
            gamma > alpha
            0 < delta < 1

    Options:
        patience (int): Used to set the maximum number of iterations to patience*(number_of_parameters+1)
        scale (double): the scale of the initial simplex, default 1.0
        alpha (double): reflection coefficient, default 1.0
        beta (double): contraction coefficient, default 0.5
        gamma (double); expansion coefficient, default 2.0
        delta (double); shrinkage coefficient, default 0.5
        adaptive_scales (boolean): if set to True we use adaptive scales instead of the default scale values.
            This sets the scales to:

            .. code-block:: python

                n = <# parameters>

                alpha = 1
                beta  = 0.75 - 1.0 / (2 * n)
                gamma = 1 + 2.0 / n
                delta = 1 - 1.0 / n

            Following the paper [1]

    References:
        [1] Gao F, Han L. Implementing the Nelder-Mead simplex algorithm with adaptive parameters.
              Comput Optim Appl. 2012;51(1):259-277. doi:10.1007/s10589-010-9329-3.
    """
    options = options or {}
    nmr_problems = x0.shape[0]
    nmr_parameters = x0.shape[1]

    penalty_data, penalty_func = _get_penalty_function(nmr_parameters, constraints_func)

    eval_func = SimpleCLFunction.from_string('''
        double evaluate(local mot_float_type* x, void* data){
            double penalty = _mle_penalty(
                x, 
                ((_nmsimplex_eval_func_data*)data)->data,
                ((_nmsimplex_eval_func_data*)data)->lower_bounds,
                ((_nmsimplex_eval_func_data*)data)->upper_bounds,
                ''' + str(options.get('penalty_weight', 1e30)) + ''',
                ((_nmsimplex_eval_func_data*)data)->penalty_data
            );

            double func_val = ''' + func.get_cl_function_name() + '''(x, ((_nmsimplex_eval_func_data*)data)->data, 0);
            
            if(isnan(func_val)){
                return INFINITY;
            }
            
            return func_val + penalty;
        }
    ''', dependencies=[func, penalty_func])

    optimizer_func = NMSimplex('evaluate', nmr_parameters, dependencies=[eval_func],
                               **_clean_options('Nelder-Mead', options))

    kernel_data = {'model_parameters': Array(x0, ctype='mot_float_type', mode='rw'),
                   'data': Struct({'data': data,
                                   'lower_bounds': lower_bounds,
                                   'upper_bounds': upper_bounds,
                                   'penalty_data': penalty_data}, '_nmsimplex_eval_func_data')}
    kernel_data.update(optimizer_func.get_kernel_data())

    return_code = optimizer_func.evaluate(
        kernel_data, nmr_problems,
        use_local_reduction=all(env.is_gpu for env in cl_runtime_info.cl_environments),
        cl_runtime_info=cl_runtime_info)

    return OptimizeResults({'x': kernel_data['model_parameters'].get_data(),
                            'status': return_code})


def _minimize_subplex(func, x0, cl_runtime_info, lower_bounds, upper_bounds,
                      constraints_func=None, data=None, options=None):
    """Variation on the Nelder-Mead Simplex method by Thomas H. Rowan.

    This method uses NMSimplex to search subspace regions for the minimum. See Rowan's thesis titled
    "Functional Stability analysis of numerical algorithms" for more details.

     The scales should satisfy the following constraints:

        .. code-block:: python

            alpha > 0
            0 < beta < 1
            gamma > 1
            gamma > alpha
            0 < delta < 1

    Options:
        patience (int): Used to set the maximum number of iterations to patience*(number_of_parameters+1)
        patience_nmsimplex (int): The maximum patience for each subspace search.
            For each subspace search we set the number of iterations to patience*(number_of_parameters_subspace+1)
        scale (double): the scale of the initial simplex, default 1.0
        alpha (double): reflection coefficient, default 1.0
        beta (double): contraction coefficient, default 0.5
        gamma (double); expansion coefficient, default 2.0
        delta (double); shrinkage coefficient, default 0.5
        psi (double): subplex specific, simplex reduction coefficient, default 0.001.
        omega (double): subplex specific, scaling reduction coefficient, default 0.01
        min_subspace_length (int): the minimum subspace length, defaults to min(2, n).
            This should hold: (1 <= min_s_d <= max_s_d <= n and min_s_d*ceil(n/nsmax_s_dmax) <= n)
        max_subspace_length (int): the maximum subspace length, defaults to min(5, n).
            This should hold: (1 <= min_s_d <= max_s_d <= n and min_s_d*ceil(n/max_s_d) <= n)

        adaptive_scales (boolean): if set to True we use adaptive scales instead of the default scale values.
            This sets the scales to:

            .. code-block:: python

                n = <# parameters>

                alpha = 1
                beta  = 0.75 - 1.0 / (2 * n)
                gamma = 1 + 2.0 / n
                delta = 1 - 1.0 / n

    References:
        [1] Gao F, Han L. Implementing the Nelder-Mead simplex algorithm with adaptive parameters.
              Comput Optim Appl. 2012;51(1):259-277. doi:10.1007/s10589-010-9329-3.
    """
    options = options or {}
    nmr_problems = x0.shape[0]
    nmr_parameters = x0.shape[1]

    penalty_data, penalty_func = _get_penalty_function(nmr_parameters, constraints_func)

    eval_func = SimpleCLFunction.from_string('''
        double evaluate(local mot_float_type* x, void* data){
            double penalty = _mle_penalty(
                x, 
                ((_subplex_eval_func_data*)data)->data,
                ((_subplex_eval_func_data*)data)->lower_bounds,
                ((_subplex_eval_func_data*)data)->upper_bounds,
                ''' + str(options.get('penalty_weight', 1e30)) + ''',
                ((_subplex_eval_func_data*)data)->penalty_data
            );

            double func_val = ''' + func.get_cl_function_name() + '''(x, ((_subplex_eval_func_data*)data)->data, 0);
            
            if(isnan(func_val)){
                return INFINITY;
            }
            
            return func_val + penalty;
        }
    ''', dependencies=[func, penalty_func])

    optimizer_func = Subplex(eval_func, nmr_parameters, **_clean_options('Subplex', options))

    kernel_data = {'model_parameters': Array(x0, ctype='mot_float_type', mode='rw'),
                   'data': Struct({'data': data,
                                   'lower_bounds': lower_bounds,
                                   'upper_bounds': upper_bounds,
                                   'penalty_data': penalty_data}, '_subplex_eval_func_data')}
    kernel_data.update(optimizer_func.get_kernel_data())

    return_code = optimizer_func.evaluate(
        kernel_data, nmr_problems,
        use_local_reduction=all(env.is_gpu for env in cl_runtime_info.cl_environments),
        cl_runtime_info=cl_runtime_info)

    return OptimizeResults({'x': kernel_data['model_parameters'].get_data(),
                            'status': return_code})


def _minimize_levenberg_marquardt(func, x0, nmr_observations, cl_runtime_info, lower_bounds, upper_bounds,
                                  constraints_func=None, data=None, options=None):
    options = options or {}
    nmr_problems = x0.shape[0]
    nmr_parameters = x0.shape[1]

    if nmr_observations < x0.shape[1]:
        raise ValueError('The number of instances per problem must be greater than the number of parameters')

    penalty_data, penalty_func = _get_penalty_function(nmr_parameters, constraints_func)

    eval_func = SimpleCLFunction.from_string('''
        void evaluate(local mot_float_type* x, void* data, local mot_float_type* result){
            double penalty = _mle_penalty(
                x, 
                ((_lm_eval_func_data*)data)->data,
                ((_lm_eval_func_data*)data)->lower_bounds,
                ((_lm_eval_func_data*)data)->upper_bounds,
                ''' + str(options.get('penalty_weight', 1e30)) + ''',
                ((_lm_eval_func_data*)data)->penalty_data
            );

            ''' + func.get_cl_function_name() + '''(x, ((_lm_eval_func_data*)data)->data, result);
            
            if(get_local_id(0) == 0){
                for(int j = 0; j < ''' + str(nmr_observations) + '''; j++){
                    result[j] += penalty;    
                }
            }
            barrier(CLK_LOCAL_MEM_FENCE);            
        }
    ''', dependencies=[func, penalty_func])

    jacobian_func = _lm_numdiff_jacobian(eval_func, nmr_parameters, nmr_observations)

    optimizer_func = LevenbergMarquardt(eval_func, nmr_parameters, nmr_observations, jacobian_func,
                                        **_clean_options('Levenberg-Marquardt', options))

    kernel_data = {'model_parameters': Array(x0, ctype='mot_float_type', mode='rw'),
                   'data': Struct({'data': data,
                                   'lower_bounds': lower_bounds,
                                   'upper_bounds': upper_bounds,
                                   'penalty_data': penalty_data,
                                   'jacobian_x_tmp': LocalMemory('mot_float_type', nmr_observations)
                                   },
                                  '_lm_eval_func_data')}
    kernel_data.update(optimizer_func.get_kernel_data())

    return_code = optimizer_func.evaluate(
        kernel_data, nmr_problems,
        use_local_reduction=all(env.is_gpu for env in cl_runtime_info.cl_environments),
        cl_runtime_info=cl_runtime_info)

    return OptimizeResults({'x': kernel_data['model_parameters'].get_data(),
                            'status': return_code})


def _lm_numdiff_jacobian(eval_func, nmr_params, nmr_observations):
    """Get a numerical differentiated Jacobian function.

    This computes the Jacobian of the observations (function vector) with respect to the parameters.

    Args:
        eval_func (mot.lib.cl_function.CLFunction): the evaluation function
        nmr_params (int): the number of parameters
        nmr_observations (int): the number of observations (the length of the function vector).

    Returns:
        mot.lib.cl_function.CLFunction: CL function for numerically estimating the Jacobian.
    """
    return SimpleCLFunction.from_string(r'''
        /**
         * Compute the Jacobian for use in the Levenberg-Marquardt optimization routine.
         *
         * This should place the output in the ``fjac`` matrix.
         *
         * Parameters:
         *
         *   model_parameters: (nmr_params,) the current point around which we want to know the Jacobian
         *   data: the current modeling data, used by the objective function
         *   fvec: (nmr_observations,), the function values corresponding to the current model parameters
         *   fjac: (nmr_parameters, nmr_observations), the memory location for the Jacobian
         */
        void _lm_numdiff_jacobian(local mot_float_type* model_parameters,
                                  void* data,
                                  local mot_float_type* fvec,
                                  local mot_float_type* const fjac){
            
            const uint nmr_params = ''' + str(nmr_params) + ''';
            const uint nmr_observations = ''' + str(nmr_observations) + ''';
            
            local mot_float_type* lower_bounds = ((_lm_eval_func_data*)data)->lower_bounds;
            local mot_float_type* upper_bounds = ((_lm_eval_func_data*)data)->upper_bounds;
            local mot_float_type* jacobian_x_tmp = ((_lm_eval_func_data*)data)->jacobian_x_tmp;
            
            mot_float_type step_size = 30 * MOT_EPSILON;
            
            for (int i = 0; i < nmr_params; i++) {
                if(model_parameters[i] + step_size > upper_bounds[i]){
                    _lm_numdiff_jacobian_backwards(model_parameters, i, step_size, data, fvec, 
                                                   fjac + i*nmr_observations, jacobian_x_tmp);
                }
                else if(model_parameters[i] - step_size < lower_bounds[i]){
                    _lm_numdiff_jacobian_forwards(model_parameters, i, step_size, data, fvec, 
                                                  fjac + i*nmr_observations, jacobian_x_tmp);
                }
                else{
                    _lm_numdiff_jacobian_centered(model_parameters, i, step_size, data, fvec,  
                                                  fjac + i*nmr_observations, jacobian_x_tmp);
                }
            }
        }
    ''', dependencies=[eval_func, SimpleCLFunction.from_string('''
        void _lm_numdiff_jacobian_centered(
                local mot_float_type* model_parameters,
                uint px,
                float step_size,
                void* data,
                local mot_float_type* fvec,
                local mot_float_type* const fjac,
                local mot_float_type* fjac_tmp){
            
            const uint nmr_observations = ''' + str(nmr_observations) + ''';
            
            uint observation_ind;
            mot_float_type temp;
            
            uint local_id = get_local_id(0);
            uint workgroup_size = get_local_size(0);
            
            // F(x + h)     
            if(get_local_id(0) == 0){
                temp = model_parameters[px];
                model_parameters[px] = temp + step_size;
            }
            barrier(CLK_LOCAL_MEM_FENCE);
            ''' + eval_func.get_cl_function_name() + '''(model_parameters, data, fjac);
            
            
            // F(x - h)
            if(get_local_id(0) == 0){
                model_parameters[px] = temp - step_size;
            }
            barrier(CLK_LOCAL_MEM_FENCE);
            ''' + eval_func.get_cl_function_name() + '''(model_parameters, data, fjac_tmp);
            
            
            // combine
            for(int i = 0; i < (nmr_observations + workgroup_size - 1) / workgroup_size; i++){
                observation_ind = i * workgroup_size + local_id;
                if(observation_ind < nmr_observations){
                    fjac[observation_ind] = (fjac[observation_ind] - fjac_tmp[observation_ind]) / (2 * step_size);
                }
            }
            
            // restore parameter vector
            if(get_local_id(0) == 0){
                model_parameters[px] = temp;
            }
            barrier(CLK_LOCAL_MEM_FENCE);
        }
    '''), SimpleCLFunction.from_string('''
        void _lm_numdiff_jacobian_backwards(
               local mot_float_type* model_parameters,
               uint px,
               float step_size,
               void* data,
               local mot_float_type* fvec,
               local mot_float_type* const fjac,
               local mot_float_type* fjac_tmp){

            const uint nmr_observations = ''' + str(nmr_observations) + ''';
            
            uint observation_ind;
            mot_float_type temp;
            
            uint local_id = get_local_id(0);
            uint workgroup_size = get_local_size(0);
            
            // F(x - h)     
            if(get_local_id(0) == 0){
               temp = model_parameters[px];
               model_parameters[px] = temp - step_size;
            }
            barrier(CLK_LOCAL_MEM_FENCE);
            ''' + eval_func.get_cl_function_name() + '''(model_parameters, data, fjac);
                        
            
            // F(x - 2*h)
            if(get_local_id(0) == 0){
               model_parameters[px] = temp - 2 * step_size;
            }
            barrier(CLK_LOCAL_MEM_FENCE);
            ''' + eval_func.get_cl_function_name() + '''(model_parameters, data, fjac_tmp);
            
            // combine
            for(int i = 0; i < (nmr_observations + workgroup_size - 1) / workgroup_size; i++){
               observation_ind = i * workgroup_size + local_id;
               if(observation_ind < nmr_observations){
                   fjac[observation_ind] = (  3 * fvec[observation_ind] 
                                            - 4 * fjac[observation_ind] 
                                            + fjac_tmp[observation_ind]) / (2 * step_size);
               }
            }
            
            // restore parameter vector
            if(get_local_id(0) == 0){
               model_parameters[px] = temp;
            }
            barrier(CLK_LOCAL_MEM_FENCE);
        }
    '''), SimpleCLFunction.from_string('''
        void _lm_numdiff_jacobian_forwards(
               local mot_float_type* model_parameters,
               uint px,
               float step_size,
               void* data,
               local mot_float_type* fvec,
               local mot_float_type* const fjac,
               local mot_float_type* fjac_tmp){

            const uint nmr_observations = ''' + str(nmr_observations) + ''';
            
            uint observation_ind;
            mot_float_type temp;
            
            uint local_id = get_local_id(0);
            uint workgroup_size = get_local_size(0);
            
            // F(x + h)     
            if(get_local_id(0) == 0){
               temp = model_parameters[px];
               model_parameters[px] = temp + step_size;
            }
            barrier(CLK_LOCAL_MEM_FENCE);
            ''' + eval_func.get_cl_function_name() + '''(model_parameters, data, fjac);
            
          
            // F(x + 2*h)
            if(get_local_id(0) == 0){
               model_parameters[px] = temp + 2 * step_size;
            }
            barrier(CLK_LOCAL_MEM_FENCE);
            ''' + eval_func.get_cl_function_name() + '''(model_parameters, data, fjac);
            
            // combine
            for(int i = 0; i < (nmr_observations + workgroup_size - 1) / workgroup_size; i++){
               observation_ind = i * workgroup_size + local_id;
               if(observation_ind < nmr_observations){
                   fjac[observation_ind] = (- 3 * fvec[observation_ind] 
                                            + 4 * fjac[observation_ind] 
                                            - fjac_tmp[observation_ind]) / (2 * step_size);
               }
            }
            
            // restore parameter vector
            if(get_local_id(0) == 0){
               model_parameters[px] = temp;
            }
            barrier(CLK_LOCAL_MEM_FENCE);
        }
    ''')])


def _get_penalty_function(nmr_parameters, constraints_func=None):
    """Get a function to compute the penalty term for the boundary conditions.

    This is meant to be used in the evaluation function of the optimization routines.

    Args:
        nmr_parameters (int): the number of parameters in the model
        constraints_func (mot.optimize.base.ConstraintFunction): function to compute (inequality) constraints.
            Should hold a CL function with the signature:

            .. code-block:: c

                void <func_name>(local const mot_float_type* const x,
                                 void* data,
                                 local mot_float_type* constraint_values);

            Where ``constraints_values`` is filled as:

            .. code-block:: c

                constraint_values[i] = g_i(x)

            That is, for each constraint function :math:`g_i`, formulated as :math:`g_i(x) <= 0`, we should return
            the function value of :math:`g_i`.

    Returns:
        tuple: Struct and SimpleCLFunction, the required data for the penalty function and the penalty function itself.
    """
    dependencies = []
    data_requirements = {'scratch': LocalMemory('double', 1)}
    constraints_code = ''

    if constraints_func and constraints_func.get_nmr_constraints() > 0:
        nmr_constraints = constraints_func.get_nmr_constraints()
        dependencies.append(constraints_func)
        data_requirements['constraints'] = LocalMemory('mot_float_type', nmr_constraints)

        constraints_code = '''
            local mot_float_type* constraints = ((_mle_penalty_data*)scratch_data)->constraints;
            
            ''' + constraints_func.get_cl_function_name() + '''(x, data, constraints);
            
            for(int i = 0; i < ''' + str(nmr_constraints) + '''; i++){
                *penalty_sum += pown(max((mot_float_type)0, constraints[i]), 2); 
            }
        '''

    data = Struct(data_requirements, '_mle_penalty_data')
    func = SimpleCLFunction.from_string('''
        double _mle_penalty(
                local mot_float_type* x,
                void* data, 
                local mot_float_type* lower_bounds, 
                local mot_float_type* upper_bounds,
                float penalty_weight,
                void* scratch_data){
            
            local double* penalty_sum = ((_mle_penalty_data*)scratch_data)->scratch;
            
            if(get_local_id(0) == 0){
                *penalty_sum = 0;
                
                // boundary conditions
                for(int i = 0; i < ''' + str(nmr_parameters) + '''; i++){
                    if(isfinite(upper_bounds[i])){
                        *penalty_sum += pown(max((mot_float_type)0, x[i] - upper_bounds[i]), 2);    
                    }
                    if(isfinite(lower_bounds[i])){
                        *penalty_sum += pown(max((mot_float_type)0, lower_bounds[i] - x[i]), 2);
                    }
                }
            }
            barrier(CLK_LOCAL_MEM_FENCE);
            
            // constraints
            ''' + constraints_code + '''
            
            return penalty_weight * *penalty_sum;
        }
    ''', dependencies=dependencies)
    return data, func




from numbers import Number
import numpy as np
from mot.lib.cl_function import SimpleCLFunction
from mot.lib.kernel_data import Array, Zeros, LocalMemory

from mot.library_functions import SimpleCLLibrary

__author__ = 'Robbert Harms'
__date__ = '2017-10-16'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


def estimate_hessian(objective_func, parameters,
                     lower_bounds=None, upper_bounds=None,
                     step_ratio=2, nmr_steps=5,
                     max_step_sizes=None,
                     data=None, cl_runtime_info=None):
    """Estimate and return the upper triangular elements of the Hessian of the given function at the given parameters.

    This calculates the Hessian using central difference (using a 2nd order Taylor expansion) with a Richardson
    extrapolation over the proposed sequence of steps. If enough steps are given, we apply a Wynn epsilon extrapolation
    on top of the Richardson extrapolated results. If more steps are left, we return the estimate with the lowest error,
    taking into account outliers using a median filter.

    The Hessian is evaluated at the steps:

    .. math::
        \quad  ((f(x + d_j e_j + d_k e_k) - f(x + d_j e_j - d_k e_k)) -
                (f(x - d_j e_j + d_k e_k) - f(x - d_j e_j - d_k e_k)) /
                (4 d_j d_k)

    where :math:`e_j` is a vector where element :math:`j` is one and the rest are zero
    and :math:`d_j` is a scalar spacing :math:`steps_j`.

    Steps are generated according to an exponentially diminishing ratio, defined as:

        steps = max_step * step_ratio**-i, i = 0,1,..,nmr_steps-1.

    Where the maximum step can be provided. For example, a maximum step of 2 with a step ratio of 2, computed for
    4 steps gives: [2.0, 1.0, 0.5, 0.25]. If lower and upper bounds are given, we use as maximum step size the largest
    step size that fits between the Hessian point and the boundaries.

    The steps define the order of the estimation, with 2 steps resulting in a O(h^2) estimate, 3 steps resulting in a
    O(h^4) estimate and 4 or more steps resulting in a O(h^6) derivative estimate.

    Args:
        objective_func (mot.lib.cl_function.CLFunction): The function we want to differentiate.
            A CL function with the signature:

            .. code-block:: c

                double <func_name>(local const mot_float_type* const x, void* data);

            The objective function has the same signature as the minimization function in MOT. For the numerical
            hessian, the ``objective_list`` parameter is ignored.

        parameters (ndarray): The parameters at which to evaluate the gradient. A (d, p) matrix with d problems,
            and p parameters
        lower_bounds (list or None): a list of length (p,) for p parameters with the lower bounds.
            Each element of the list can be a scalar or a vector (of the same length as the number
            of problem instances). To disable bounds for this parameter use -np.inf.
        upper_bounds (list or None): a list of length (p,) for p parameters with the upper bounds.
            Each element of the list can be a scalar or a vector (of the same length as the number
            of problem instances). To disable bounds for this parameter use np.inf.
        step_ratio (float): the ratio at which the steps diminish.
        nmr_steps (int): the number of steps we will generate. We will calculate the derivative for each of these
            step sizes and extrapolate the best step size from among them. The minimum number of steps is 1.
        max_step_sizes (float or ndarray or None): the maximum step size, or the maximum step size per parameter.
            If None is given, we use 0.1 for all parameters. If a float is given, we use that for all parameters.
            If a list is given, it should be of the same length as the number of parameters.
        data (mot.lib.kernel_data.KernelData): the user provided data for the ``void* data`` pointer.
        cl_runtime_info (mot.configuration.CLRuntimeInfo): the runtime information

    Returns:
        ndarray: per problem instance a vector with the upper triangular elements of the Hessian matrix.
            This array can hold NaN's, for elements where the Hessian failed to approximate.
    """
    if len(parameters.shape) == 1:
        parameters = parameters[None, :]

    nmr_voxels = parameters.shape[0]
    nmr_params = parameters.shape[1]
    nmr_derivatives = nmr_params * (nmr_params + 1) // 2

    initial_step = _get_initial_step(parameters, lower_bounds, upper_bounds, max_step_sizes)

    kernel_data = {
        'parameters': Array(parameters, ctype='mot_float_type'),
        'initial_step': Array(initial_step, ctype='float'),
        'derivatives': Zeros((nmr_voxels, nmr_derivatives), 'double'),
        'errors': Zeros((nmr_voxels, nmr_derivatives), 'double'),
        'x_tmp': LocalMemory('mot_float_type', nmr_params),
        'data': data,
        'scratch': LocalMemory('double', nmr_steps + (nmr_steps - 1) + nmr_steps)
    }

    hessian_kernel = SimpleCLFunction.from_string('''
        void _numdiff_hessian(
                global mot_float_type* parameters,
                global float* initial_step,
                global double* derivatives,
                global double* errors,
                local mot_float_type* x_tmp,
                void* data,
                local double* scratch){

            if(get_local_id(0) == 0){
                for(uint i = 0; i < ''' + str(nmr_params) + '''; i++){
                    x_tmp[i] = parameters[i];
                }
            }
            barrier(CLK_LOCAL_MEM_FENCE);

            double f_x_input = ''' + objective_func.get_cl_function_name() + '''(x_tmp, data);

            // upper triangle loop
            uint coord_ind = 0;
            for(int i = 0; i < ''' + str(nmr_params) + '''; i++){
                for(int j = i; j < ''' + str(nmr_params) + '''; j++){
                    _numdiff_hessian_element(
                        data, x_tmp, f_x_input, i, j, initial_step, 
                        derivatives + coord_ind, errors + coord_ind, scratch);

                    coord_ind++;
                }
            }
        }
    ''', dependencies=[objective_func,
                       _get_numdiff_hessian_element_func(objective_func, nmr_steps, step_ratio)])

    hessian_kernel.evaluate(kernel_data, nmr_voxels, use_local_reduction=True, cl_runtime_info=cl_runtime_info)

    return kernel_data['derivatives'].get_data()


def _get_numdiff_hessian_element_func(objective_func, nmr_steps, step_ratio):
    """Return a function to compute one element of the Hessian matrix."""
    return SimpleCLFunction.from_string('''
        /**
         * Compute the Hessian using (possibly) multiple steps with various interpolations. 
         */ 
        void _numdiff_hessian_element(
                void* data, local mot_float_type* x_tmp, mot_float_type f_x_input,
                uint px, uint py, global float* initial_step, global double* derivative, 
                global double* error, local double* scratch){

            const uint nmr_steps = ''' + str(nmr_steps) + ''';
            uint nmr_steps_remaining = nmr_steps;

            local double* scratch_ind = scratch;
            local double* steps = scratch_ind;      scratch_ind += nmr_steps; 
            local double* errors = scratch_ind;     scratch_ind += nmr_steps - 1;
            local double* steps_tmp = scratch_ind;  scratch_ind += nmr_steps;
            
            if(get_local_id(0) == 0){
                for(int i = 0; i < nmr_steps - 1; i++){
                    errors[i] = 0;
                }
            }
            barrier(CLK_LOCAL_MEM_FENCE);
            
            _numdiff_hessian_steps(data, x_tmp, f_x_input, px, py, steps, initial_step);

            if(nmr_steps_remaining > 1){
                nmr_steps_remaining = _numdiff_hessian_richardson_extrapolation(steps); 
                barrier(CLK_LOCAL_MEM_FENCE);
            }
            
            if(nmr_steps_remaining >= 3){
                nmr_steps_remaining = _numdiff_wynn_extrapolation(steps, errors, nmr_steps_remaining);
                barrier(CLK_LOCAL_MEM_FENCE);                
            }
            
            if(nmr_steps_remaining > 1){
                _numdiff_find_best_step(steps, errors, steps_tmp, nmr_steps_remaining);
                barrier(CLK_LOCAL_MEM_FENCE);
            }
            
            if(get_local_id(0) == 0){
                *derivative = steps[0];
                *error = errors[0];
            }
        }
    ''', dependencies=[
        _get_numdiff_hessian_steps_func(objective_func, nmr_steps, step_ratio),
        _get_numdiff_hessian_richardson_extrapolation_func(nmr_steps, step_ratio),
        _get_numdiff_wynn_extrapolation_func(),
        _get_numdiff_find_best_step_func()
    ])


def _get_numdiff_hessian_steps_func(objective_func, nmr_steps, step_ratio):
    """Get a function to compute the multiple step sizes for a single element of the Hessian."""
    return SimpleCLFunction.from_string('''
        /**
         * Compute one element of the Hessian for a number of steps.
         * 
         * This uses the initial steps in the data structure, indexed by the parameters to change (px, py).
         *
         * Args:
         *  data: the data container
         *  x_tmp: the array with the input parameters, needs to be writable, although it will return
         *         the same values.
         *  f_x_input: the objective function value at the original set of parameters  
         *  px: the index of the first parameter to perturbate
         *  py: the index of the second parameter to perturbate
         *  steps: storage location for the output steps
         *  initial_step: the initial steps, array of same length as x_temp
         */
        void _numdiff_hessian_steps(void* data, local mot_float_type* x_tmp, 
                                    mot_float_type f_x_input,
                                    uint px, uint py, 
                                    local double* steps, 
                                    global float* initial_step){

            double step_x;
            double step_y;
            double tmp;
            bool is_first_workitem = get_local_id(0) == 0;

            if(px == py){
                for(uint step_ind = 0; step_ind < ''' + str(nmr_steps) + '''; step_ind++){
                    step_x = initial_step[px] / pown(''' + str(float(step_ratio)) + ''', step_ind);

                    tmp = (
                          _numdiff_hessian_eval_step_mono(data, x_tmp, px, 2 * step_x)
                        + _numdiff_hessian_eval_step_mono(data, x_tmp, px, -2 * step_x)
                        - 2 * f_x_input
                    ) / (4 * step_x * step_x);    

                    if(is_first_workitem){
                        steps[step_ind] = tmp;
                    }
                }
            }
            else{
                for(uint step_ind = 0; step_ind < ''' + str(nmr_steps) + '''; step_ind++){
                    step_x = initial_step[px] / pown(''' + str(float(step_ratio)) + ''', step_ind);
                    step_y = initial_step[py] / pown(''' + str(float(step_ratio)) + ''', step_ind);

                    tmp = (
                          _numdiff_hessian_eval_step_bi(data, x_tmp, px, step_x, py, step_y)
                        - _numdiff_hessian_eval_step_bi(data, x_tmp, px, step_x, py, -step_y)
                        - _numdiff_hessian_eval_step_bi(data, x_tmp, px, -step_x, py, step_y)
                        + _numdiff_hessian_eval_step_bi(data, x_tmp, px, -step_x, py, -step_y)
                    ) / (4 * step_x * step_y);

                    if(is_first_workitem){
                        steps[step_ind] = tmp;
                    }                       
                }
            }
        }
    ''', dependencies=[SimpleCLFunction.from_string('''
        /**
         * Evaluate the model with a perturbation in one dimensions.
         *
         * Args:
         *  data: the data container
         *  x_tmp: the array with the input parameters, needs to be writable, although it will return
         *         the same values.
         *  perturb_dim0: the index (into the x_tmp parameters) of the parameter to perturbate
         *  perturb_0: the added perturbation of the index corresponding to ``perturb_dim_0``
         *
         * Returns:
         *  the function evaluated at the parameters plus their perturbation.
         */
        double _numdiff_hessian_eval_step_mono(
                void* data, local mot_float_type* x_tmp, 
                uint perturb_dim_0, mot_float_type perturb_0){

            mot_float_type old_0;
            double return_val;

            if(get_local_id(0) == 0){
                old_0 = x_tmp[perturb_dim_0];
                x_tmp[perturb_dim_0] += perturb_0;
            }
            barrier(CLK_LOCAL_MEM_FENCE);

            return_val = ''' + objective_func.get_cl_function_name() + '''(x_tmp, data);
            barrier(CLK_LOCAL_MEM_FENCE);

            if(get_local_id(0) == 0){
                x_tmp[perturb_dim_0] = old_0;
            }
            barrier(CLK_LOCAL_MEM_FENCE);

            return return_val;
        }
    '''), SimpleCLFunction.from_string('''
        /**
         * Evaluate the model with a perturbation in two dimensions.
         *
         * Args:
         *  data: the data container
         *  x_tmp: the array with the input parameters, needs to be writable, although it will return
         *         the same values.
         *  perturb_dim_0: the index (into the x_tmp parameters) of the first parameter to perturbate
         *  perturb_0: the added perturbation of the index corresponding to ``perturb_dim_0``
         *  perturb_dim_1: the index (into the x_tmp parameters) of the second parameter to perturbate
         *  perturb_1: the added perturbation of the index corresponding to ``perturb_dim_1``
         *
         * Returns:
         *  the function evaluated at the parameters plus their perturbation.
         */
        double _numdiff_hessian_eval_step_bi(
                void* data, local mot_float_type* x_tmp, 
                uint perturb_dim_0, mot_float_type perturb_0,
                uint perturb_dim_1, mot_float_type perturb_1){

            mot_float_type old_0;
            mot_float_type old_1;
            double return_val;

            if(get_local_id(0) == 0){
                old_0 = x_tmp[perturb_dim_0];
                old_1 = x_tmp[perturb_dim_1];

                x_tmp[perturb_dim_0] += perturb_0;
                x_tmp[perturb_dim_1] += perturb_1;
            }
            barrier(CLK_LOCAL_MEM_FENCE);

            return_val = ''' + objective_func.get_cl_function_name() + '''(x_tmp, data);
            barrier(CLK_LOCAL_MEM_FENCE);

            if(get_local_id(0) == 0){
                x_tmp[perturb_dim_0] = old_0;
                x_tmp[perturb_dim_1] = old_1;
            }
            barrier(CLK_LOCAL_MEM_FENCE);

            return return_val;
        }
    ''')])


def _get_numdiff_hessian_richardson_extrapolation_func(nmr_steps, step_ratio):
    return SimpleCLFunction.from_string('''
        /**
         * Apply the Richardson extrapolation to the derivatives computed with multiple steps.
         * 
         * Having for every problem instance and every Hessian element multiple derivatives computed 
         * with decreasing steps, we can now apply Richardson extrapolation to reduce the error term from O(h^2) to
         * O(h^4) or O(h^6) depending on how many steps we have calculated.
         *
         * This method only considers extrapolation up to the sixth error order. For a set of two derivatives we compute
         * a single fourth order approximation, for three derivatives or more, we compute ``n-2`` sixth order 
         * approximations. 
         * 
         * Args:
         *  steps: on input, the steps we are convoluting with Richardson extrapolation. On output, 
         *         the convoluted steps, these are less than the input, see the return value.
         * 
         * Returns:
         *  nmr_steps: the number of steps remaining after convolution
         */
        uint _numdiff_hessian_richardson_extrapolation(local double* steps){
            
            const uint nmr_steps = ''' + str(nmr_steps) + ''';
            uint nmr_steps_remaining = nmr_steps;
            
            if(get_local_id(0) == 0){
                // 4th order approximations
                for(uint i = 0; i < nmr_steps - 1; i++){
                    steps[i] = richardson_extrapolate(steps[i], steps[i + 1], ''' + str(step_ratio) + ''', 2);
                }
                nmr_steps_remaining--;
                
                // 6th order approximations
                for(uint i = 0; i < nmr_steps - 2; i++){
                    steps[i] = richardson_extrapolate(steps[i], steps[i + 1], ''' + str(step_ratio) + ''', 4);
                }
                nmr_steps_remaining--;
            }

            return nmr_steps_remaining;
        }
    ''', dependencies=[richardson_extrapolation()])


def _get_numdiff_find_best_step_func():
    return SimpleCLFunction.from_string('''
        /**
         * Find the step with the lowest error.
         *
         * This will first apply a median filter on the steps to locate derivatives which deviate strongly 
         * from the distribution of the steps. These deviations are added as extra error term to the error term from 
         * the extrapolations. Afterwards it chooses the derivative with the lowest error. 
         *
         * Args:
         *  steps: on input, array with the input steps. On output, a single step with the best derivative
         *  errors: on input, array with the error terms per step, on output the error for the returned step
         *  scratch: temporary array of the same length as the steps array
         *  nmr_steps: the number of steps in the input array
         */
        void _numdiff_find_best_step(local double* steps, local double* errors, local double* scratch, uint nmr_steps){
            if(get_local_id(0) == 0){
                bool all_nan = true;
                for(int i = 0; i < nmr_steps; i++){
                    if(isnan(steps[i])){
                        steps[i] = 0;
                    }
                    else{
                        all_nan = false;
                    }
                }
    
                if(all_nan){
                    steps[0] = NAN;
                    errors[0] = NAN;
    
                    return;
                }
    
                _get_median_outlier_error(steps, scratch, nmr_steps);
                
                double lowest_error = INFINITY;
                int ind_lowest_error = 0; 
                for(int i = 0; i < nmr_steps; i++){
                    if(errors[i] + scratch[i] < lowest_error){
                        lowest_error = errors[i] + scratch[i];
                        ind_lowest_error = i;
                    } 
                }
    
                steps[0] = steps[ind_lowest_error];
                errors[0] = errors[ind_lowest_error] + scratch[ind_lowest_error];    
            }
        }
    ''', dependencies=[SimpleCLFunction.from_string('''
        /**
         * This function tries to detect outliers by means of an median filter. 
         * If a value is a factor of 10 to 1 in either direction away from the median, we weight it 
         * with an additional error.
         *
         * This uses a hardcoded trim factor of 10 to 1 as relative distance.
         * 
         * Args:
         *  steps: 
         * 
         */
        void _get_median_outlier_error(local double* steps, local double* median_errors, uint nmr_steps){
            float trim_factor = 10;   
            int i;

            local double* sorted_values = median_errors;
            for(int i = 0; i < nmr_steps; i++){
                sorted_values[i] = steps[i];
            }            
            _numdiff_sort_values(sorted_values, nmr_steps);

            double p25 = _numdiff_get_percentile(sorted_values, nmr_steps, 0.25);
            double p50 = _numdiff_get_percentile(sorted_values, nmr_steps, 0.50);
            double p75 = _numdiff_get_percentile(sorted_values, nmr_steps, 0.75);

            double iqr = fabs(p75 - p25);
            double abs_median = fabs(p50);

            bool is_outlier;
            for(i = 0; i < nmr_steps; i++){
                is_outlier = (((fabs(steps[i]) < (abs_median / trim_factor)) +
                               (fabs(steps[i]) > (abs_median * trim_factor))) * (abs_median > 1e-8) +
                              ((steps[i] < p25 - 1.5 * iqr) + (p75 + 1.5 * iqr < steps[i])));

                median_errors[i] = is_outlier * fabs(steps[i] - p50);
            }            
        }

    ''', dependencies=[SimpleCLFunction.from_string('''
            /** 
             * Uses bubblesort to sort the given values in ascending order.
             * 
             */
            void _numdiff_sort_values(local double* values, int n){  
                int i, j;
                double tmp;

                for(i = 0; i < n; i++){
                    for(j = 0; j < n - i - 1; j++){
                        if(values[j] > values[j + 1]){
                            tmp = values[j];
                            values[j] = values[j + 1];
                            values[j + 1] = tmp;    
                        }
                    }
                }
            }
        '''), SimpleCLFunction.from_string('''
            /**
             * Find the distribution value for the requested percentile.
             * 
             * This uses linear interpolation with C == 1, as defined in 
             * https://en.wikipedia.org/wiki/Percentile#The_nearest-rank_method.
             * 
             * Args:
             *  sorted_values: array of sorted values
             *  nmr_values: number of values
             *  percentile: the requested percentile, between 0 and 1.
             * 
             * Returns:
             *  the value for the percentile
             */
            double _numdiff_get_percentile(local double* sorted_values, int nmr_values, double percentile){
                double rank = percentile * (nmr_values - 1);
                int rank_int = (int) rank;
                double fraction = rank - rank_int;
                return sorted_values[rank_int] + fraction * (sorted_values[rank_int + 1] - sorted_values[rank_int]);    
            }
        ''')])])


def _get_numdiff_wynn_extrapolation_func():
    return SimpleCLFunction.from_string('''
        /**
         * Apply Wynn extrapolation to the derivatives.
         *
         * This algorithm, known in the Python Numdifftools as DEA3, attempts to extrapolate non-linearly to a better
         * estimate of the sequence's limiting value, thus improving the rate of convergence. The routine is based on the
         * epsilon algorithm of P. Wynn [1].
         *
         * References:
         *   - [1] C. Brezinski and M. Redivo Zaglia (1991)
         *           "Extrapolation Methods. Theory and Practice", North-Holland.
         *
         *   - [2] C. Brezinski (1977)
         *           "Acceleration de la convergence en analyse numerique",
         *           "Lecture Notes in Math.", vol. 584,
         *           Springer-Verlag, New York, 1977.
         * 
         *   - [3] E. J. Weniger (1989)
         *           "Nonlinear sequence transformations for the acceleration of
         *           convergence and the summation of divergent series"
         *           Computer Physics Reports Vol. 10, 189 - 371
         *           http://arxiv.org/abs/math/0306302v1
         * 
         * Args:
         *  steps: on input, array with the input steps. On output, the extrapolated steps
         *  errors: the errors of the extrapolated steps
         *  nmr_steps: the number of steps in the input array
         * 
         * Returns:
         *  the number of steps in the output 
         */
        uint _numdiff_wynn_extrapolation(local double* steps, local double* errors, uint nmr_steps){
            if(get_local_id(0) == 0){
                double v0, v1, v2; 

                double delta0, delta1; 
                double err0, err1; 
                double tol0, tol1; 

                double ss;
                bool converged;

                double result;

                for(uint i = 0; i < nmr_steps - 2; i++){
                    v0 = steps[i];
                    v1 = steps[i + 1];
                    v2 = steps[i + 2];

                    delta0 = v1 - v0;
                    delta1 = v2 - v1;

                    err0 = fabs(delta0);
                    err1 = fabs(delta1);

                    tol0 = max(fabs(v0), fabs(v1)) * MOT_EPSILON;
                    tol1 = max(fabs(v1), fabs(v2)) * MOT_EPSILON;

                    // avoid division by zero and overflow
                    if(err0 < MOT_MIN){
                        delta0 = MOT_MIN;
                    }
                    if(err1 < MOT_MIN){
                        delta1 = MOT_MIN;
                    }

                    ss = 1.0 / delta1 - 1.0 / delta0 + MOT_MIN;
                    converged = ((err0 <= tol0) && (err1 <= tol1)) || (fabs(ss * v1) <= 1e-3);

                    result = (converged ? v2 : v1 + 1.0/ss);

                    steps[i] = result;
                    errors[i] = err0 + err1 + (converged ? tol1 * 10 : fabs(result - v2));
                }
            }
            return nmr_steps - 2;
        }
    ''', dependencies=[])


def _get_initial_step(parameters, lower_bounds, upper_bounds, max_step_sizes):
    """Get an initial step size to use for every parameter.

    This chooses the step sizes based on the maximum step size and the lower and upper bounds.

    Args:
        parameters (ndarray): The parameters at which to evaluate the gradient. A (d, p) matrix with d problems,
            p parameters and n samples.
        lower_bounds (list): lower bounds
        upper_bounds (list): upper bounds
        max_step_sizes (list or None): the maximum step size, or the maximum step size per parameter. Defaults to 0.1

    Returns:
        ndarray: for every problem instance the vector with the initial step size for each parameter.
    """
    nmr_params = parameters.shape[1]

    initial_step = np.zeros_like(parameters)

    if max_step_sizes is None:
        max_step_sizes = 0.1
    if isinstance(max_step_sizes, Number):
        max_step_sizes = [max_step_sizes] * nmr_params
    max_step_sizes = np.array(max_step_sizes)

    for ind in range(parameters.shape[1]):
        minimum_allowed_step = np.minimum(np.abs(parameters[:, ind] - lower_bounds[ind]),
                                          np.abs(upper_bounds[ind] - parameters[:, ind]))
        initial_step[:, ind] = np.minimum(minimum_allowed_step, max_step_sizes[ind])

    return initial_step / 2.


class richardson_extrapolation(SimpleCLLibrary):

    def __init__(self):
        """Get the Richardson extrapolation function."""
        super().__init__('''
            /**
             * Apply a Richardson extrapolation on the given input.
             *
             * See https://en.wikipedia.org/wiki/Richardson_extrapolation. 
             *
             * Args:
             *  v1: the primary value at a step size of h
             *  v2: the secondary value, computed at a step size of h/step_ratio 
             *  step_ratio: the ratio at which the steps diminish
             *  approximation_order: the order of the current approximations
             * 
             * Returns:
             *   the Richardson extrapolation
             */
            double richardson_extrapolate(double v1, double v2, float step_ratio, uint approximation_order){
                double div = 1 / (pown(step_ratio, approximation_order) - 1);
                return (div + 1) * v2 - div * v1;
            }
        ''')



"""Contains the runtime configuration of MOT.

This consists of two parts, functions to get the current runtime settings and configuration actions to update these
settings. To set a new configuration, create a new :py:class:`ConfigAction` and use this within a context environment
using :py:func:`config_context`. Example:

.. code-block:: python

    from mot.configuration import RuntimeConfigurationAction, config_context

    with config_context(RuntimeConfigurationAction(...)):
        ...

"""
from contextlib import contextmanager
from copy import copy
import numpy as np
from .lib.cl_environments import CLEnvironmentFactory

__author__ = 'Robbert Harms'
__date__ = "2015-07-22"
__maintainer__ = "Robbert Harms"
__email__ = "robbert.harms@maastrichtuniversity.nl"


"""The runtime configuration, this can be overwritten at run time.

For any of the AbstractCLRoutines it holds that if no suitable defaults are given we use the ones provided by this
module. This entire module acts as a singleton containing the current runtime configuration.
"""
_config = {
    'cl_environments': CLEnvironmentFactory.smart_device_selection(preferred_device_type='GPU'),
    'compile_flags': ['-cl-denorms-are-zero', '-cl-mad-enable', '-cl-no-signed-zeros'],
    'double_precision': False
}


def get_cl_environments():
    """Get the current CL environment to use during CL calculations.

    Returns:
        list of CLEnvironment: the current list of CL environments.
    """
    return _config['cl_environments']


def set_cl_environments(cl_environments):
    """Set the current CL environments to the given list

    Please note that this will change the global configuration, i.e. this is a persistent change. If you do not want
    a persistent state change, consider using :func:`~mot.configuration.config_context` instead.

    Args:
        cl_environments (list of CLEnvironment): the new list of CL environments.

    Raises:
        ValueError: if the list of environments is empty
    """
    if not cl_environments:
        raise ValueError('The list of CL Environments is empty.')
    _config['cl_environments'] = cl_environments


def get_compile_flags():
    """Get the default compile flags to use in a CL routine.

    Returns:
        list: the default list of compile flags we wish to use
    """
    return list(_config['compile_flags'])


def set_compile_flags(compile_flags):
    """Set the current compile flags.

    Args:
        compile_flags (list): the new list of compile flags
    """
    _config['compile_flags'] = compile_flags


def set_default_proposal_update(proposal_update):
    """Set the default proposal update function to use in sample.

    Please note that this will change the global configuration, i.e. this is a persistent change. If you do not want
    a persistent state change, consider using :func:`~mot.configuration.config_context` instead.

    Args:
        mot.model_building.parameter_functions.proposal_updates.ProposalUpdate: the new proposal update function
            to use by default if no specific one is provided.
    """
    _config['default_proposal_update'] = proposal_update


def use_double_precision():
    """Check if we run the computations in default precision or not.

    Returns:
        boolean: if we run the computations in double precision or not
    """
    return _config['double_precision']


def set_use_double_precision(double_precision):
    """Set the default use of double precision.

    Returns:
        boolean: if we use double precision by default or not
    """
    _config['double_precision'] = double_precision


@contextmanager
def config_context(config_action):
    """Creates a context in which the config action is applied and unapplies the configuration after execution.

    Args:
        config_action (ConfigAction): the configuration action to use
    """
    config_action.apply()
    yield
    config_action.unapply()


class ConfigAction:

    def __init__(self):
        """Defines a configuration action for use in a configuration context.

        This should define an apply and unapply function that sets and unsets the configuration options.

        The applying action needs to remember the state before the application of the action.
        """

    def apply(self):
        """Apply the current action to the current runtime configuration."""

    def unapply(self):
        """Reset the current configuration to the previous state."""


class SimpleConfigAction(ConfigAction):

    def __init__(self):
        """Defines a default implementation of a configuration action.

        This simple config implements a default ``apply()`` method that saves the current state and a default
        ``unapply()`` that restores the previous state.

        For developers, it is easiest to implement ``_apply()`` such that you do not manually need to store the old
        configuraration.
        """
        super().__init__()
        self._old_config = {}

    def apply(self):
        """Apply the current action to the current runtime configuration."""
        self._old_config = {k: v for k, v in _config.items()}
        self._apply()

    def unapply(self):
        """Reset the current configuration to the previous state."""
        for key, value in self._old_config.items():
            _config[key] = value

    def _apply(self):
        """Implement this function add apply() logic after this class saves the current config."""


class CLRuntimeAction(SimpleConfigAction):

    def __init__(self, cl_runtime_info):
        """Set the current configuration to use the information in the given configuration action.

        Args:
            cl_runtime_info (CLRuntimeInfo): the runtime info with the configuration options
        """
        super().__init__()
        self._cl_runtime_info = cl_runtime_info

    def _apply(self):
        set_cl_environments(self._cl_runtime_info.cl_environments)
        set_compile_flags(self._cl_runtime_info._compile_flags)
        set_use_double_precision(self._cl_runtime_info.double_precision)


class RuntimeConfigurationAction(SimpleConfigAction):

    def __init__(self, cl_environments=None, compile_flags=None, double_precision=None):
        """Updates the runtime settings.

        Args:
            cl_environments (list of CLEnvironment): the new CL environments we wish to use for future computations
            compile_flags (list): the list of compile flags to use during analysis.
            double_precision (boolean): if we compute in double precision or not
        """
        super().__init__()
        self._cl_environments = cl_environments
        self._compile_flags = compile_flags
        self._double_precision = double_precision

    def _apply(self):
        if self._cl_environments is not None:
            set_cl_environments(self._cl_environments)

        if self._compile_flags is not None:
            set_compile_flags(self._compile_flags)

        if self._double_precision is not None:
            set_use_double_precision(self._double_precision)


class VoidConfigurationAction(ConfigAction):

    def __init__(self):
        """Does nothing, useful as a default config action.
        """
        super().__init__()


class CLRuntimeInfo:

    def __init__(self, cl_environments=None, compile_flags=None, double_precision=None):
        """All information necessary for applying operations using OpenCL.

        Args:
            cl_environments (list of mot.lib.cl_environments.CLEnvironment): The list of CL environments used by
                this routine. If None is given we use the defaults in the current configuration.
            compile_flags (list): the list of compile flags to use during analysis.
            double_precision (boolean): if we apply the computations in double precision or in single float precision.
                By default we go for single float precision.
        """
        self._cl_environments = cl_environments
        self._compile_flags = compile_flags
        self._double_precision = double_precision

        if self._cl_environments is None:
            self._cl_environments = get_cl_environments()

        if self._compile_flags is None:
            self._compile_flags = get_compile_flags()

        if self._double_precision is None:
            self._double_precision = use_double_precision()

    @property
    def cl_environments(self):
        return self._cl_environments

    @property
    def mot_float_dtype(self):
        if self._double_precision:
            return np.float64
        return np.float32

    @property
    def double_precision(self):
        return self._double_precision

    @property
    def compile_flags(self):
        """Get all defined compile flags."""
        return self._compile_flags



"""Several routines from the EISPACK-C code.

All routines are prefixed with 'eispack_' for use in MOT.

Reference:
    https://people.sc.fsu.edu/~jburkardt/c_src/eispack/eispack.html .
"""
from mot.library_functions import SimpleCLLibrary

__author__ = 'Robbert Harms'
__date__ = '2018-12-08'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class eispack_tred2(SimpleCLLibrary):
    def __init__(self):
        super().__init__('''
void eispack_tred2 ( int n, double* a, double* d, double* e, double* z ){
/*
  Purpose:

    TRED2 transforms a real symmetric matrix to symmetric tridiagonal form.

  Discussion:

    TRED2 reduces a real symmetric matrix to a
    symmetric tridiagonal matrix using and accumulating
    orthogonal similarity transformations.

    A and Z may coincide, in which case a single storage area is used
    for the input of A and the output of Z.

  Licensing:

    This code is distributed under the GNU LGPL license.

  Modified:

    03 November 2012

  Author:

    Original FORTRAN77 version by Smith, Boyle, Dongarra, Garbow, Ikebe,
    Klema, Moler.
    C version by John Burkardt.

  Reference:

    Martin, Reinsch, Wilkinson,
    TRED2,
    Numerische Mathematik,
    Volume 11, pages 181-195, 1968.

    James Wilkinson, Christian Reinsch,
    Handbook for Automatic Computation,
    Volume II, Linear Algebra, Part 2,
    Springer, 1971,
    ISBN: 0387054146,
    LC: QA251.W67.

    Brian Smith, James Boyle, Jack Dongarra, Burton Garbow,
    Yasuhiko Ikebe, Virginia Klema, Cleve Moler,
    Matrix Eigensystem Routines, EISPACK Guide,
    Lecture Notes in Computer Science, Volume 6,
    Springer Verlag, 1976,
    ISBN13: 978-3540075462,
    LC: QA193.M37.

  Parameters:

    Input, int N, the order of the matrix.

    Input, double A[N*N], the real symmetric input matrix.  Only the
    lower triangle of the matrix need be supplied.

    Output, double D[N], the diagonal elements of the tridiagonal
    matrix.

    Output, double E[N], contains the subdiagonal elements of the
    tridiagonal matrix in E(2:N).  E(1) is set to zero.

    Output, double Z[N*N], the orthogonal transformation matrix
    produced in the reduction.
*/
  double f;
  double g;
  double h;
  double hh;
  int i;
  int j;
  int k;
  int l;
  double scale;

  for ( j = 0; j < n; j++ )
  {
    for ( i = j; i < n; i++ )
    {
      z[i+j*n] = a[i+j*n];
    }
  }

  for ( j = 0; j < n; j++ )
  {
    d[j] = a[n-1+j*n];
  }

  for ( i = n - 1; 1 <= i; i-- )
  {
    l = i - 1;
    h = 0.0;
/*
  Scale row.
*/
    scale = 0.0;
    for ( k = 0; k <= l; k++ )
    {
      scale = scale + fabs ( d[k] );
    }

    if ( scale == 0.0 )
    {
      e[i] = d[l];

      for ( j = 0; j <= l; j++ )
      {
        d[j]     = z[l+j*n];
        z[i+j*n] = 0.0;
        z[j+i*n] = 0.0;
      }
      d[i] = 0.0;
      continue;
    }

    for ( k = 0; k <= l; k++ )
    {
      d[k] = d[k] / scale;
    }

    h = 0.0;
    for ( k = 0; k <= l; k++ )
    {
      h = h + d[k] * d[k];
    }

    f = d[l];
    g = - sqrt ( h ) * sign ( f );
    e[i] = scale * g;
    h = h - f * g;
    d[l] = f - g;
/*
  Form A*U.
*/
    for ( k = 0; k <= l; k++ )
    {
      e[k] = 0.0;
    }

    for ( j = 0; j <= l; j++ )
    {
      f = d[j];
      z[j+i*n] = f;
      g = e[j] + z[j+j*n] * f;

      for ( k = j + 1; k <= l; k++ )
      {
        g = g + z[k+j*n] * d[k];
        e[k] = e[k] + z[k+j*n] * f;
      }
      e[j] = g;
    }
/*
  Form P.
*/
    for ( k = 0; k <= l; k++ )
    {
      e[k] = e[k] / h;
    }
    f = 0.0;
    for ( k = 0; k <= l; k++ )
    {
      f = f + e[k] * d[k];
    }
    hh = 0.5 * f / h;
/*
  Form Q.
*/
    for ( k = 0; k <= l; k++ )
    {
      e[k] = e[k] - hh * d[k];
    }
/*
  Form reduced A.
*/
    for ( j = 0; j <= l; j++ )
    {
      f = d[j];
      g = e[j];

      for ( k = j; k <= l; k++ )
      {
        z[k+j*n] = z[k+j*n] - f * e[k] - g * d[k];
      }
      d[j] = z[l+j*n];
      z[i+j*n] = 0.0;
    }
    d[i] = h;
  }
/*
  Accumulation of transformation matrices.
*/
  for ( i = 1; i < n; i++ )
  {
    l = i - 1;
    z[n-1+l*n] = z[l+l*n];
    z[l+l*n] = 1.0;
    h = d[i];

    if ( h != 0.0 )
    {
      for ( k = 0; k <= l; k++ )
      {
        d[k] = z[k+i*n] / h;
      }
      for ( j = 0; j <= l; j++ )
      {
        g = 0.0;
        for ( k = 0; k <= l; k++ )
        {
          g = g + z[k+i*n] * z[k+j*n];
        }
        for ( k = 0; k <= l; k++ )
        {
          z[k+j*n] = z[k+j*n] - g * d[k];
        }
      }
    }
    for ( k = 0; k <= l; k++ )
    {
      z[k+i*n] = 0.0;
    }
  }

  for ( j = 0; j < n; j++ )
  {
    d[j] = z[n-1+j*n];
  }

  for ( j = 0; j < n - 1; j++ )
  {
    z[n-1+j*n] = 0.0;
  }
  z[n-1+(n-1)*n] = 1.0;

  e[0] = 0.0;

  return;
}

        ''')


class eispack_tql2(SimpleCLLibrary):
    def __init__(self):
        super().__init__('''
int eispack_tql2 ( int n, double* d, double* e, double* z ){
/*
  Purpose:

    TQL2 computes all eigenvalues/vectors, real symmetric tridiagonal matrix.

  Discussion:

    TQL2 finds the eigenvalues and eigenvectors of a symmetric
    tridiagonal matrix by the QL method.  The eigenvectors of a full
    symmetric matrix can also be found if TRED2 has been used to reduce this
    full matrix to tridiagonal form.

  Licensing:

    This code is distributed under the GNU LGPL license.

  Modified:

    08 November 2012

  Author:

    Original FORTRAN77 version by Smith, Boyle, Dongarra, Garbow, Ikebe,
    Klema, Moler.
    C version by John Burkardt.

  Reference:

    Bowdler, Martin, Reinsch, Wilkinson,
    TQL2,
    Numerische Mathematik,
    Volume 11, pages 293-306, 1968.

    James Wilkinson, Christian Reinsch,
    Handbook for Automatic Computation,
    Volume II, Linear Algebra, Part 2,
    Springer, 1971,
    ISBN: 0387054146,
    LC: QA251.W67.

    Brian Smith, James Boyle, Jack Dongarra, Burton Garbow,
    Yasuhiko Ikebe, Virginia Klema, Cleve Moler,
    Matrix Eigensystem Routines, EISPACK Guide,
    Lecture Notes in Computer Science, Volume 6,
    Springer Verlag, 1976,
    ISBN13: 978-3540075462,
    LC: QA193.M37.

  Parameters:

    Input, int N, the order of the matrix.

    Input/output, double D[N].  On input, the diagonal elements of
    the matrix.  On output, the eigenvalues in ascending order.  If an error
    exit is made, the eigenvalues are correct but unordered for indices
    1,2,...,IERR-1.

    Input/output, double E[N].  On input, E(2:N) contains the
    subdiagonal elements of the input matrix, and E(1) is arbitrary.
    On output, E has been destroyed.

    Input, double Z[N*N].  On input, the transformation matrix
    produced in the reduction by TRED2, if performed.  If the eigenvectors of
    the tridiagonal matrix are desired, Z must contain the identity matrix.
    On output, Z contains the orthonormal eigenvectors of the symmetric
    tridiagonal (or full) matrix.  If an error exit is made, Z contains
    the eigenvectors associated with the stored eigenvalues.

    Output, int TQL2, error flag.
    0, normal return,
    J, if the J-th eigenvalue has not been determined after
    30 iterations.
*/
  double c;
  double c2;
  double c3;
  double dl1;
  double el1;
  double f;
  double g;
  double h;
  int i;
  int ierr;
  int ii;
  int j;
  int k;
  int l;
  int l1;
  int l2;
  int m;
  int mml;
  double p;
  double r;
  double s;
  double s2;
  double t;
  double tst1;
  double tst2;

  ierr = 0;

  if ( n == 1 )
  {
    return ierr;
  }

  for ( i = 1; i < n; i++ )
  {
    e[i-1] = e[i];
  }

  f = 0.0;
  tst1 = 0.0;
  e[n-1] = 0.0;

  for ( l = 0; l < n; l++ )
  {
    j = 0;
    h = fabs ( d[l] ) + fabs ( e[l] );
    tst1 = fmax ( tst1, h );
/*
  Look for a small sub-diagonal element.
*/
    for ( m = l; m < n; m++ )
    {
      tst2 = tst1 + fabs ( e[m] );
      if ( tst2 == tst1 )
      {
        break;
      }
    }

    if ( m != l )
    {
      for ( ; ; )
      {
        if ( 30 <= j )
        {
          ierr = l + 1;
          return ierr;
        }

        j = j + 1;
/*
  Form shift.
*/
        l1 = l + 1;
        l2 = l1 + 1;
        g = d[l];
        p = ( d[l1] - g ) / ( 2.0 * e[l] );
        r = hypot ( p, 1.0 );
        d[l] = e[l] / ( p + sign ( p ) * fabs ( r ) );
        d[l1] = e[l] * ( p + sign ( p ) * fabs ( r ) );
        dl1 = d[l1];
        h = g - d[l];
        for ( i = l2; i < n; i++ )
        {
          d[i] = d[i] - h;
        }
        f = f + h;
/*
  QL transformation.
*/
        p = d[m];
        c = 1.0;
        c2 = c;
        el1 = e[l1];
        s = 0.0;
        mml = m - l;

        for ( ii = 1; ii <= mml; ii++ )
        {
          c3 = c2;
          c2 = c;
          s2 = s;
          i = m - ii;
          g = c * e[i];
          h = c * p;
          r = hypot ( p, e[i] );
          e[i+1] = s * r;
          s = e[i] / r;
          c = p / r;
          p = c * d[i] - s * g;
          d[i+1] = h + s * ( c * g + s * d[i] );
/*
  Form vector.
*/
          for ( k = 0; k < n; k++ )
          {
            h = z[k+(i+1)*n];
            z[k+(i+1)*n] = s * z[k+i*n] + c * h;
            z[k+i*n] = c * z[k+i*n] - s * h;
          }
        }
        p = - s * s2 * c3 * el1 * e[l] / dl1;
        e[l] = s * p;
        d[l] = c * p;
        tst2 = tst1 + fabs ( e[l] );

        if ( tst2 <= tst1 )
        {
          break;
        }
      }
    }
    d[l] = d[l] + f;
  }
/*
  Order eigenvalues and eigenvectors.
*/
  for ( ii = 1; ii < n; ii++ )
  {
    i = ii - 1;
    k = i;
    p = d[i];
    for ( j = ii; j < n; j++ )
    {
      if ( d[j] < p )
      {
        k = j;
        p = d[j];
      }
    }

    if ( k != i )
    {
      d[k] = d[i];
      d[i] = p;
      for ( j = 0; j < n; j++ )
      {
        t        = z[j+i*n];
        z[j+i*n] = z[j+k*n];
        z[j+k*n] = t;
      }
    }
  }
  return ierr;
}
        ''')



import numpy as np

from mot.lib.cl_function import SimpleCLFunction
from mot.lib.kernel_data import Struct, Array, LocalMemory
from mot.lib.utils import parse_cl_function
from mot.sample.base import AbstractSampler

__author__ = 'Robbert Harms'
__date__ = '2018-10-20'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class ThoughtfulWalk(AbstractSampler):

    def __init__(self, ll_func, log_prior_func, x0, x1, finalize_proposal_func=None,
                 subset_size=4, walk_scale=1.5, traverse_scale=6,
                 move_probabilities=(60/122, 60/122., 1/122., 1/122.), **kwargs):
        """The thoughtful- or traverse- walk MCMC algorithm.

        The t-walk (twalk) algorithm of Christen and Fox [1] is a general-purpose MCMC algorithm that
        requires no tuning, is scale-invariant, is technically non-adaptive (but self-adjusting), and can sample
        from target distributions with arbitrary scale and correlation structures. A random subset of one of two
        vectors is moved around the state-space to influence one of two chains, per iteration.

        Args:
            ll_func (mot.lib.cl_function.CLFunction): The log-likelihood function.
            log_prior_func (mot.lib.cl_function.CLFunction): The log-prior function.
            x0 (ndarray): the starting positions for the sampler. Should be a two dimensional matrix
                with for every modeling instance (first dimension) and every parameter (second dimension) a value.
            x1 (ndarray): the second coordinate vector for every model instance. This should be of the same size as
                ``x0``, but then with different values. This basically defines the initial sampling space.
            subset_size (int): the expected number of parameters to move in one step.
                This is variable n_1 in the article [1]. If there are less parameters than the subset_size, we will
                use all parameters.
            walk_scale (float): the scalar of the walk move. This is parameter a_w in [1]. A reasonable range
                given by [1] is [0.3, 2]
            traverse_scale (float): the scalar of the traverse move. This is parameter a_t in [1]. A reasonable range
                given by [1] is [2, 10]
            move_probabilities (Tuple[float]): the probabilities of the four different moves supported in the
                t-walk algorithm. Defaults to (0.4918, 0.4918, 0.0082, 0.0082) [1], with respectively the probability
                of walk, traverse, hop, blow moves.
            finalize_proposal_func (mot.lib.cl_function.CLFunction): a CL function to finalize every proposal by
                the sampling routine. This allows the model to change a proposal before computing the
                prior or likelihood probabilities. If None, we will not use this callback.

                As an example, suppose you are sampling a polar coordinate :math:`\theta` defined on
                :math:`[0, 2\pi]` with a random walk Metropolis proposal distribution. This distribution might propose
                positions outside of the range of :math:`\theta`. Of course the model function could deal with that by
                taking the modulus of the input, but then you have to post-process the chain with the same
                transformation. Instead, this function allows changing the proposal before it is put into the model
                and before it will be stored in the chain.

                This function should return a proposal that is equivalent (but not necessarily equal)
                to the provided proposal.

                Signature:

                .. code-block:: c

                    void <func_name>(void* data, local mot_float_type* x);

        References:
            [1] Christen JA, Foxy C. A general purpose sampling algorithm for continuous distributions (the t-walk).
                Bayesian Anal. 2010;5(2):263-282. doi:10.1214/10-BA603.
        """
        super().__init__(ll_func, log_prior_func, x0, **kwargs)

        self._subset_size = subset_size
        self._param_choose_prob = min(self._nmr_params, self._subset_size) / (float(self._nmr_params))
        self._walk_scale = walk_scale
        self._traverse_scale = traverse_scale
        self._move_probabilities = move_probabilities

        float_type = self._cl_runtime_info.mot_float_dtype
        if x1.shape != x0.shape:
            self._x1 = np.require(np.tile(x1, (x0.shape[0], 1)), requirements='CAOW', dtype=float_type)
        else:
            self._x1 = np.require(np.copy(x1), requirements='CAOW', dtype=float_type)

        self._x1_log_likelihood = np.zeros(self._nmr_problems, dtype=float_type)
        self._x1_log_prior = np.zeros(self._nmr_problems, dtype=float_type)

        self._finalize_proposal_func = finalize_proposal_func or SimpleCLFunction.from_string(
            'void finalizeProposal(void* data, local mot_float_type* x){}')

        self._initialize_likelihood_prior(self._x1, self._x1_log_likelihood, self._x1_log_prior)

    def _get_mcmc_method_kernel_data(self):
        return Struct({
            'x1_position': Array(self._x1, 'mot_float_type', mode='rw', ensure_zero_copy=True),
            'x1_log_likelihood': Array(self._x1_log_likelihood, 'mot_float_type', mode='rw', ensure_zero_copy=True),
            'x1_log_prior': Array(self._x1_log_prior, 'mot_float_type', mode='rw', ensure_zero_copy=True),
            'scratch_mft': LocalMemory('mot_float_type', self._nmr_params + 2),
            'scratch_int': LocalMemory('int', self._nmr_params + 4),
        }, '_twalk_data')

    def _get_state_update_cl_func(self, nmr_samples, thinning, return_output):
        func = parse_cl_function('''    
            void _twalk_advance_chain(
                    void* method_data,
                    void* data,
                    ulong current_iteration, 
                    void* rng_data,
                    int chain_ind,
                    global mot_float_type* current_position,
                    global mot_float_type* current_log_likelihood,
                    global mot_float_type* current_log_prior){
                
                _twalk_data* twalk_data = (_twalk_data*)method_data;
                bool is_first_work_item = get_local_id(0) == 0;
                
                local int* kernel_ind = twalk_data->scratch_int + 1;
                local int* nmr_params_selected = twalk_data->scratch_int + 2;
                local int* proposal_accepted = twalk_data->scratch_int + 3;
                local int* params_selector = twalk_data->scratch_int + 4;
                
                local mot_float_type* proposal_ll = twalk_data->scratch_mft;
                local mot_float_type* proposal_lprior = twalk_data->scratch_mft + 1;
                local mot_float_type* proposal = twalk_data->scratch_mft + 2;
                
                *proposal_accepted = false;
                *proposal_ll = 0;
                *proposal_lprior = 0;
                
                global mot_float_type* main_chain;
                global mot_float_type* helper_chain;
                global mot_float_type* main_ll;
                global mot_float_type* main_lprior;
                
                if(is_first_work_item){
                    float r = frand(rng_data);
                    
                    if(r < ''' + str(self._move_probabilities[0]) + '''){
                        *kernel_ind = 0;
                    }
                    else if(r < ''' + str(sum(self._move_probabilities[:2])) + '''){
                        *kernel_ind = 1;
                    }
                    else if(r < ''' + str(sum(self._move_probabilities[:3])) + '''){
                        *kernel_ind = 2;
                    }
                    else{
                        *kernel_ind = 3;
                    }
                    
                    *nmr_params_selected = 0;
                    for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                        params_selector[i] = frand(rng_data) < ''' + str(self._param_choose_prob) + ''';
                        (*nmr_params_selected)++;
                    }
                }
                barrier(CLK_LOCAL_MEM_FENCE);
                
                if(chain_ind == 0){
                    main_chain = twalk_data->x1_position;
                    helper_chain = current_position;
                    main_ll = twalk_data->x1_log_likelihood;
                    main_lprior = twalk_data->x1_log_prior;
                }
                else{
                    main_chain = current_position;
                    helper_chain = twalk_data->x1_position;    
                    main_ll = current_log_likelihood;
                    main_lprior = current_log_prior;
                }
            
                if(*kernel_ind == 0){
                    _twalk_walk_move(main_chain, helper_chain, main_ll, main_lprior,
                                     proposal, proposal_ll, proposal_lprior, proposal_accepted,
                                     params_selector, *nmr_params_selected, data, rng_data);
                }
                else if(*kernel_ind == 1){
                    _twalk_traverse_move(main_chain, helper_chain, main_ll, main_lprior,
                                         proposal, proposal_ll, proposal_lprior, proposal_accepted,
                                         params_selector, *nmr_params_selected, data, rng_data);
                }
                else if(*kernel_ind == 2){
                    _twalk_hop_move(main_chain, helper_chain, main_ll, main_lprior,
                                    proposal, proposal_ll, proposal_lprior, proposal_accepted,
                                    params_selector, *nmr_params_selected, data, rng_data);
                }
                else{
                    _twalk_blow_move(main_chain, helper_chain, main_ll, main_lprior,
                                     proposal, proposal_ll, proposal_lprior, proposal_accepted,
                                     params_selector, *nmr_params_selected, data, rng_data);
                }
                
                if(is_first_work_item && *proposal_accepted){
                    if(chain_ind == 0){
                        for(uint k = 0; k < ''' + str(self._nmr_params) + '''; k++){
                            twalk_data->x1_position[k] = proposal[k];
                        }
                        *twalk_data->x1_log_likelihood = *proposal_ll;
                        *twalk_data->x1_log_prior = *proposal_lprior;
                    }
                    else{
                        for(uint k = 0; k < ''' + str(self._nmr_params) + '''; k++){
                            current_position[k] = proposal[k];
                        }
                        *current_log_likelihood = *proposal_ll;
                        *current_log_prior = *proposal_lprior;
                    }
                }
                barrier(CLK_LOCAL_MEM_FENCE);
            }
            
            void _advanceSampler(
                    void* method_data,
                    void* data,
                    ulong current_iteration, 
                    void* rng_data,
                    global mot_float_type* current_position,
                    global mot_float_type* current_log_likelihood,
                    global mot_float_type* current_log_prior){
                
                _twalk_data* twalk_data = (_twalk_data*)method_data;
                
                local int* chain_ind = twalk_data->scratch_int;
                *chain_ind = 0;
                
                while(*chain_ind != 1){
                    if(get_local_id(0) == 0){
                        *chain_ind = frand(rng_data) > 0.5;
                    }
                    barrier(CLK_LOCAL_MEM_FENCE);
                    
                    _twalk_advance_chain(method_data, data, current_iteration, rng_data, *chain_ind, 
                                         current_position, current_log_likelihood, current_log_prior);                         
                }
            }
            
        ''', dependencies=[self._finalize_proposal_func, self._get_walk_move(), self._get_traverse_move(),
                           self._get_hop_move(), self._get_blow_move()])
        return func.get_cl_code()

    def _get_walk_move(self):
        return parse_cl_function('''
            void _twalk_walk_move_proposal(
                    global mot_float_type* main_chain, 
                    global mot_float_type* helper_chain,
                    local mot_float_type* proposal,
                    local int* params_selector,
                    void* rng_data){

                float u, a;
                const float aw = ''' + str(self._walk_scale) + ''';

                for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    if(params_selector[i]){
                        u = frand(rng_data);
                        a = (aw / (1 + aw)) * (-1 + 2 * u + aw * u * u);

                        proposal[i] = main_chain[i] + a * (main_chain[i] - helper_chain[i]);
                    }
                    else{
                        proposal[i] = main_chain[i];
                    }
                }
            }

            void _twalk_walk_move(
                    global mot_float_type* main_chain, 
                    global mot_float_type* helper_chain,
                    global mot_float_type* main_ll, 
                    global mot_float_type* main_lprior,
                    local mot_float_type* proposal,
                    local mot_float_type* proposal_ll, 
                    local mot_float_type* proposal_lprior,
                    local int* proposal_accepted,
                    local int* params_selector,
                    int nmr_params_selected,
                    void* data,
                    void* rng_data){

                bool is_first_work_item = get_local_id(0) == 0;

                if(is_first_work_item){
                    _twalk_walk_move_proposal(main_chain, helper_chain, proposal, params_selector, rng_data);
                    ''' + self._finalize_proposal_func.get_cl_function_name() + '''(data, proposal);

                    *proposal_lprior = _computeLogPrior(proposal, data);
                }
                barrier(CLK_LOCAL_MEM_FENCE);

                if(exp(*proposal_lprior) > 0){
                    *proposal_ll = _computeLogLikelihood(proposal, data);

                    if(is_first_work_item){    
                        *proposal_accepted = frand(rng_data) < exp((*proposal_ll + *proposal_lprior) 
                                                                   - (*main_ll + *main_lprior));
                    }
                }
                barrier(CLK_LOCAL_MEM_FENCE);
            }    
        ''')

    def _get_traverse_move(self):
        return parse_cl_function('''
            float _twalk_traverse_move_compute_beta(void* rng_data){            
                float4 r = frand4(rng_data);
                float at = ''' + str(self._traverse_scale) + ''';
                
                if(r.x < (at - 1.0) / (2.0 * at)){
                    return exp(1.0 / (at + 1.0) * log(r.y));
                }
                return exp(1.0 / (1.0 - at) * log(r.y));
            }
            
            void _twalk_traverse_move_proposal(
                    global mot_float_type* main_chain, 
                    global mot_float_type* helper_chain,
                    local mot_float_type* proposal,
                    local int* params_selector,
                    mot_float_type beta,
                    void* rng_data){
                             
                for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    if(params_selector[i]){
                        proposal[i] = helper_chain[i] + beta * (helper_chain[i] - main_chain[i]);
                    }
                    else{
                        proposal[i] = main_chain[i];
                    }
                }
            }
            
            void _twalk_traverse_move(
                    global mot_float_type* main_chain, 
                    global mot_float_type* helper_chain,
                    global mot_float_type* main_ll, 
                    global mot_float_type* main_lprior,
                    local mot_float_type* proposal,
                    local mot_float_type* proposal_ll, 
                    local mot_float_type* proposal_lprior,
                    local int* proposal_accepted,
                    local int* params_selector,
                    int nmr_params_selected,
                    void* data,
                    void* rng_data){
                
                float beta;
                bool is_first_work_item = get_local_id(0) == 0;
                
                if(is_first_work_item){
                    beta = _twalk_traverse_move_compute_beta(rng_data);    
                    _twalk_traverse_move_proposal(main_chain, helper_chain, proposal, 
                                                  params_selector, beta, rng_data);
                    ''' + self._finalize_proposal_func.get_cl_function_name() + '''(data, proposal);
                    
                    *proposal_lprior = _computeLogPrior(proposal, data);
                }
                barrier(CLK_LOCAL_MEM_FENCE);
                
                if(exp(*proposal_lprior) > 0){
                    *proposal_ll = _computeLogLikelihood(proposal, data);
                    
                    if(is_first_work_item){
                        if(nmr_params_selected == 0){
                            *proposal_accepted = true;
                        }
                        else{
                            *proposal_accepted = frand(rng_data) < exp((*proposal_ll + *proposal_lprior) 
                                                                       - (*main_ll + *main_lprior)
                                                                       + (nmr_params_selected - 2) * log(beta));
                        }
                    }
                }
                barrier(CLK_LOCAL_MEM_FENCE);
            }
        ''')

    def _get_hop_move(self):
        return parse_cl_function('''
            void _twalk_hop_move_proposal(
                    global mot_float_type* main_chain, 
                    global mot_float_type* helper_chain,
                    local mot_float_type* proposal,
                    local int* params_selector,
                    void* rng_data){
                
                mot_float_type sigma = 0;
                for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    sigma = max(sigma, params_selector[i] * fabs(main_chain[i] - helper_chain[i]));
                }
                sigma /= 3.0;
                    
                for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    if(params_selector[i]){
                        proposal[i] = main_chain[i] + sigma * frandn(rng_data);
                    }
                    else{
                        proposal[i] = main_chain[i];
                    }
                }
            }
            
            float _hop_move_hasting_criteria_xy(
                    global mot_float_type* proposal,
                    local mot_float_type* main_chain,
                    global mot_float_type* helper_chain,
                    local int* params_selector,
                    int nmr_params_selected){
                
                mot_float_type sigma = 0;
                double sum = 0;
                for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    sigma = max(sigma, params_selector[i] * fabs(main_chain[i] - helper_chain[i]));
                    sum += pown(proposal[i] - main_chain[i], 2);
                }
                sigma /= 3.0;
                    
                if(nmr_params_selected > 0){
                    return -(nmr_params_selected/2.0) * log(2*M_PI) 
                           + nmr_params_selected * log(3.0)
                           - nmr_params_selected * log(sigma) 
                           - 9 * sum / (2 * sigma * sigma);
                }
                return 0;
            }    
            
            
            float _hop_move_hasting_criteria_yx(
                    local mot_float_type* proposal,
                    global mot_float_type* main_chain,
                    global mot_float_type* helper_chain,
                    local int* params_selector,
                    int nmr_params_selected){
                
                mot_float_type sigma = 0;
                double sum = 0;
                for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    sigma = max(sigma, params_selector[i] * fabs(main_chain[i] - helper_chain[i]));
                    sum += pown(proposal[i] - main_chain[i], 2);
                }
                sigma /= 3.0;
                
                if(nmr_params_selected > 0){
                    return -(nmr_params_selected/2.0) * log(2*M_PI) 
                           + nmr_params_selected * log(3.0)
                           - nmr_params_selected * log(sigma) 
                           - 9 * sum / (2 * sigma * sigma);
                }
                return 0;
            }    
            
            void _twalk_hop_move(
                    global mot_float_type* main_chain, 
                    global mot_float_type* helper_chain,
                    global mot_float_type* main_ll, 
                    global mot_float_type* main_lprior,
                    local mot_float_type* proposal,
                    local mot_float_type* proposal_ll, 
                    local mot_float_type* proposal_lprior,
                    local int* proposal_accepted,
                    local int* params_selector,
                    int nmr_params_selected,
                    void* data,
                    void* rng_data){
            
                bool is_first_work_item = get_local_id(0) == 0;

                if(is_first_work_item){    
                    _twalk_hop_move_proposal(main_chain, helper_chain, proposal, params_selector, rng_data);
                    ''' + self._finalize_proposal_func.get_cl_function_name() + '''(data, proposal);

                    *proposal_lprior = _computeLogPrior(proposal, data);
                }
                barrier(CLK_LOCAL_MEM_FENCE);

                if(exp(*proposal_lprior) > 0){
                    *proposal_ll = _computeLogLikelihood(proposal, data);

                    if(is_first_work_item){
                        float g_xy = _hop_move_hasting_criteria_xy(main_chain, proposal, helper_chain, 
                                                                   params_selector, nmr_params_selected);
                        float g_yx = _hop_move_hasting_criteria_yx(proposal, main_chain, helper_chain, 
                                                                   params_selector, nmr_params_selected);
                        
                        *proposal_accepted = frand(rng_data) < exp((*proposal_ll + *proposal_lprior) 
                                                                       - (*main_ll + *main_lprior) 
                                                                       + (g_xy - g_yx));
                    }
                }
                barrier(CLK_LOCAL_MEM_FENCE);
            }
        ''')

    def _get_blow_move(self):
        return parse_cl_function('''
            void _twalk_blow_move_proposal(
                    global mot_float_type* main_chain, 
                    global mot_float_type* helper_chain,
                    local mot_float_type* proposal,
                    local int* params_selector,
                    void* rng_data){

                mot_float_type sigma = 0;
                for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    sigma = max(sigma, params_selector[i] * fabs(main_chain[i] - helper_chain[i]));
                }

                for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    if(params_selector[i]){
                        proposal[i] = helper_chain[i] + sigma * frandn(rng_data);
                    }
                    else{
                        proposal[i] = main_chain[i];
                    }
                }
            }

            float _blow_move_hasting_criteria_xy(
                    global mot_float_type* proposal,
                    local mot_float_type* main_chain,
                    global mot_float_type* helper_chain,
                    local int* params_selector,
                    int nmr_params_selected){
                
                mot_float_type sigma = 0;
                double sum = 0;
                for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    sigma = max(sigma, params_selector[i] * fabs(main_chain[i] - helper_chain[i]));
                    sum += pown(proposal[i] - helper_chain[i], 2);
                }
                
                if(nmr_params_selected > 0){
                    return -(nmr_params_selected/2.0) * log(2*M_PI) 
                           - nmr_params_selected * log(sigma) 
                           - sum / (2 * sigma * sigma);
                }
                return 0;
            }    

            float _blow_move_hasting_criteria_yx(
                    local mot_float_type* proposal,
                    global mot_float_type* main_chain,
                    global mot_float_type* helper_chain,
                    local int* params_selector,
                    int nmr_params_selected){
                
                mot_float_type sigma = 0;
                double sum = 0;
                for(uint i = 0; i < ''' + str(self._nmr_params) + '''; i++){
                    sigma = max(sigma, params_selector[i] * fabs(main_chain[i] - helper_chain[i]));
                    sum += pown(proposal[i] - helper_chain[i], 2);
                }
                
                if(nmr_params_selected > 0){
                    return -(nmr_params_selected/2.0) * log(2*M_PI) 
                           - nmr_params_selected * log(sigma) 
                           - sum / (2 * sigma * sigma);
                }
                return 0;
            }    

            void _twalk_blow_move(
                    global mot_float_type* main_chain, 
                    global mot_float_type* helper_chain,
                    global mot_float_type* main_ll, 
                    global mot_float_type* main_lprior,
                    local mot_float_type* proposal,
                    local mot_float_type* proposal_ll, 
                    local mot_float_type* proposal_lprior,
                    local int* proposal_accepted,
                    local int* params_selector,
                    int nmr_params_selected,
                    void* data,
                    void* rng_data){

                bool is_first_work_item = get_local_id(0) == 0;

                if(is_first_work_item){    
                    _twalk_blow_move_proposal(main_chain, helper_chain, proposal, params_selector, rng_data);
                    ''' + self._finalize_proposal_func.get_cl_function_name() + '''(data, proposal);

                    *proposal_lprior = _computeLogPrior(proposal, data);
                }
                barrier(CLK_LOCAL_MEM_FENCE);

                if(exp(*proposal_lprior) > 0){
                    *proposal_ll = _computeLogLikelihood(proposal, data);

                    if(is_first_work_item){
                        float g_xy = _blow_move_hasting_criteria_xy(main_chain, proposal, helper_chain, 
                                                                    params_selector, nmr_params_selected);
                        float g_yx = _blow_move_hasting_criteria_yx(proposal, main_chain, helper_chain, 
                                                                    params_selector, nmr_params_selected);

                        *proposal_accepted = frand(rng_data) < exp((*proposal_ll + *proposal_lprior) 
                                                                       - (*main_ll + *main_lprior) 
                                                                       + (g_xy - g_yx));
                    }
                }
                barrier(CLK_LOCAL_MEM_FENCE);
            }
        ''')



import os

from mot.lib.cl_function import SimpleCLCodeObject
from mot.lib.kernel_data import LocalMemory
from mot.library_functions.base import SimpleCLLibrary, SimpleCLLibraryFromFile, CLLibrary
from pkg_resources import resource_filename

from mot.library_functions.eispack import eispack_tred2, eispack_tql2
from mot.library_functions.unity import log1pmx
from mot.library_functions.polynomials import p1evl, polevl, ratevl, real_zeros_cubic_pol
from mot.library_functions.continuous_distributions.normal import normal_cdf, normal_pdf, normal_logpdf, normal_ppf
from mot.library_functions.continuous_distributions.gamma import gamma_pdf, gamma_logpdf, gamma_ppf, gamma_cdf, \
    gamma_cdf_approx, gamma_ppf_approx
from mot.library_functions.error_functions import dawson, CerfImWOfX, erfi
from mot.library_functions.legendre_polynomial import FirstLegendreTerm, LegendreTerms, \
    EvenLegendreTerms, OddLegendreTerms


__author__ = 'Robbert Harms'
__date__ = '2018-05-07'
__maintainer__ = 'Robbert Harms'
__email__ = 'robbert.harms@maastrichtuniversity.nl'
__licence__ = 'LGPL v3'


class Besseli0(SimpleCLLibrary):
    def __init__(self):
        """Return the zeroth-order modified Bessel function of the first kind

        Original author of C code: M.G.R. Vogelaar
        """
        super().__init__('''
            double bessel_i0(double x){
                double y;
    
                if(fabs(x) < 3.75){
                    y = (x / 3.75) * (x / 3.75);                  
    
                    return 1.0 + y * (3.5156229 
                                      + y * (3.0899424
                                             + y * (1.2067492 
                                                    + y * (0.2659732
                                                           + y * (0.360768e-1 
                                                                  + y * 0.45813e-2)))));
                }
    
                y = 3.75 / fabs(x);
                return (exp(fabs(x)) / sqrt(fabs(x))) 
                        * (0.39894228
                           + y * (0.1328592e-1
                                  + y * (0.225319e-2
                                         + y * (-0.157565e-2
                                                + y * (0.916281e-2
                                                       + y * (-0.2057706e-1
                                                              + y * (0.2635537e-1
                                                                     + y * (-0.1647633e-1
                                                                            + y * 0.392377e-2))))))));
            }
        ''')


class LogBesseli0(SimpleCLLibrary):
    def __init__(self):
        """Return the log of the zeroth-order modified Bessel function of the first kind."""
        super().__init__('''
            double log_bessel_i0(double x){
                if(x < 700){
                  return log(bessel_i0(x));
                }
                return x - log(2.0 * M_PI * x)/2.0;
            }
        ''', dependencies=(Besseli0(),))


class LogCosh(SimpleCLLibrary):
    def __init__(self):
        """Computes :math:`log(cosh(x))`

        For large x this will try to estimate it without overflow. For small x we use the opencl functions log and cos.
        The estimation for large numbers has been taken from:
        https://github.com/JaneliaSciComp/tmt/blob/master/basics/logcosh.m
        """
        super().__init__('''
            double log_cosh(double x){
                if(x < 50){
                    return log(cosh(x));
                }
                return fabs(x) + log(1 + exp(-2.0 * fabs(x))) - log(2.0);
            }
        ''')


class Rand123(SimpleCLCodeObject):
    def __init__(self):
        generator = 'threefry'

        src = open(os.path.abspath(resource_filename('mot', 'data/opencl/random123/openclfeatures.h'), ), 'r').read()
        src += open(os.path.abspath(resource_filename('mot', 'data/opencl/random123/array.h'), ), 'r').read()
        src += open(os.path.abspath(resource_filename('mot', 'data/opencl/random123/{}.h'.format(generator)), ),
                    'r').read()
        src += (open(os.path.abspath(resource_filename('mot', 'data/opencl/random123/rand123.h'), ), 'r').read() % {
            'GENERATOR_NAME': (generator)
        })
        super().__init__(src)


class EuclidianNormFunction(SimpleCLLibraryFromFile):
    def __init__(self, memspace='private', memtype='mot_float_type'):
        """A CL functions for calculating the Euclidian distance between n values.

        Args:
            memspace (str): The memory space of the memtyped array (private, constant, global).
            memtype (str): the memory type to use, double, float, mot_float_type, ...
        """
        super().__init__(
            memtype,
            self.__class__.__name__ + '_' + memspace + '_' + memtype,
            [],
            resource_filename('mot', 'data/opencl/euclidian_norm.cl'),
            var_replace_dict={'MEMSPACE': memspace, 'MEMTYPE': memtype})


class simpsons_rule(SimpleCLLibrary):

    def __init__(self, function_name):
        """Create a CL function for integrating a function using Simpson's rule.

        This creates a CL function specifically meant for integrating the function of the given name.
        The name of the generated CL function will be 'simpsons_rule_<function_name>'.

        Args:
            function_name (str): the name of the function to integrate, accepting the arguments:
                - a: the lower bound of the integral
                - b: the upper bound of the integral
                - n: the number of steps, i.e. the number of approximations to make
                - data: a pointer to some data, this is passed on to the function we are integrating.
        """
        super().__init__('''
            double simpsons_rule_{f}(double a, double b, uint n, void* data){{
                double h = (b - a) / n;

                double sum_odds = {f}(a + h/2.0, data);
                double sum_evens = 0.0;

                double x;

                for(uint i = 1; i < n; i++){{
                    sum_odds += {f}(a + h * i + h / 2.0, data);
                    sum_evens += {f}(a + h * i, data);
                }}

                return h / 6.0 * ({f}(a, data) + {f}(b, data) + 4.0 * sum_odds + 2.0 * sum_evens);
            }}
        '''.format(f=function_name))


class linear_cubic_interpolation(SimpleCLLibrary):

    def __init__(self):
        """Cubic interpolation for a one-dimensional grid.

        This uses the theory of Cubic Hermite splines for interpolating a one-dimensional grid of values.

        At the borders, it will clip the values to the nearest border.

        For more information on this method, see https://en.wikipedia.org/wiki/Cubic_Hermite_spline.

        Example usage:
            constant float data[] = {1.0, 2.0, 5.0, 6.0};
            linear_cubic_interpolation(1.5, 4, data);
        """
        super().__init__('''
            double linear_cubic_interpolation(double x, int y_len, constant float* y_values){
                int n = x;
                double u = x - n;

                double p0 = y_values[min(max((int)0, n - 1), y_len - 1)];
                double p1 = y_values[min(max((int)0, n    ), y_len - 1)];
                double p2 = y_values[min(max((int)0, n + 1), y_len - 1)];
                double p3 = y_values[min(max((int)0, n + 2), y_len - 1)];

                double a = 0.5 * (-p0 + 3.0 * p1 - 3.0 * p2 + p3);
                double b = 0.5 * (2.0 * p0 - 5.0 * p1 + 4.0 * p2 - p3);
                double c = 0.5 * (-p0 + p2);
                double d = p1;

                return d + u * (c + u * (b + u * a));                
            }
        ''')


class eigenvalues_3x3_symmetric(SimpleCLLibrary):

    def __init__(self):
        """Calculate the eigenvalues of a symmetric 3x3 matrix.

        This simple algorithm only works in case of a real and symmetric matrix.

        The input to this function is an array with the upper triangular elements of the matrix.
        The output are the eigenvalues such that eig1 >= eig2 >= eig3, i.e. from large to small.

        Args:
            A: the upper triangle of an 3x3 matrix
            v: the output eigenvalues as a vector of three elements.

        References:
            [1]: https://en.wikipedia.org/wiki/Eigenvalue_algorithm#3.C3.973_matrices
            [2]: Smith, Oliver K. (April 1961), "Eigenvalues of a symmetric 3 × 3 matrix.", Communications of the ACM,
                 4 (4): 168, doi:10.1145/355578.366316
        """
        super().__init__('''
            void eigenvalues_3x3_symmetric(double* A, double* v){
                double p1 = (A[1] * A[1]) + (A[2] * A[2]) + (A[4] * A[4]);
        
                if (p1 == 0.0){
                    v[0] = A[0];
                    v[1] = A[3];
                    v[2] = A[5];
                    return;
                }
        
                double q = (A[0] + A[3] + A[5]) / 3.0;
                double p = sqrt(((A[0] - q)*(A[0] - q) + (A[3] - q)*(A[3] - q) + (A[5] - q)*(A[5] - q) + 2*p1) / 6.0);
        
                double r = (
                     ((A[0] - q)/p) * ((((A[3] - q)/p) * ((A[5] - q)/p)) - ((A[4]/p) * (A[4]/p))) 
                    - (A[1]/p)      * ((A[1]/p) * ((A[5] - q)/p) - (A[2]/p) * (A[4]/p)) 
                    + (A[2]/p)      * ((A[1]/p) * (A[4]/p) - (A[2]/p) * ((A[3] - q)/p))
                ) / 2.0;
        
                double phi;
                if(r <= -1){
                    phi = M_PI / 3.0;
                }
                else if(r >= 1){
                    phi = 0;
                }
                else{
                    phi = acos(r) / 3;
                }
        
                v[0] = q + 2 * p * cos(phi);
                v[2] = q + 2 * p * cos(phi + (2*M_PI/3.0));
                v[1] = 3 * q - v[0] - v[2];            
            }
        ''')


class multiply_square_matrices(SimpleCLLibrary):
    def __init__(self):
        """Matrix multiplication of two square matrices.

        Having this as a special function is slightly faster than a more generic matrix multiplication algorithm.

        All matrices are expected to be in c/row-major order.

        Parameters:
            n: the rectangular size of the matrix (the number of rows / the number of columns).
            A[n*n]: the left matrix
            B[n*n]: the right matrix
            C[n*n]: the output matrix
        """
        super().__init__('''
            void multiply_square_matrices(uint n, double* A, double* B, double* C){
                int i, j, z;
                double sum;
                for(int ind = 0; ind < n*n; ind++){
                    i = ind / n;
                    j = ind - n * i;

                    sum = 0;
                    for(z = 0; z < n; z++){
                        sum += A[i * n + z] * B[z * n + j];
                    }
                    C[ind] = sum;
                }
            }
        ''')


class eigen_decompose_real_symmetric_matrix(SimpleCLLibrary):

    def __init__(self):
        """Computes eigenvalues and eigenvectors of real symmetric matrix.

        This uses the RS routine from the EISPACK code, to be found at:
        https://people.sc.fsu.edu/~jburkardt/c_src/eispack/eispack.html

        It first tri-diagonalizes the matrix using Householder transformations and then computes the diagonal
        using QR transformations.

        This routine only works with real symmetric matrices as input.

        Parameters:
            n: input, the order of the matrix.
            A[n*n]: input, the real symmetric matrix to invert
            W[n]: output, the eigenvalues in ascending order.
            Z[n*n]: output, the eigenvectors, can coincide with A[n*n].
            scratch[n]: input, scratch data

        Output:
            error code: the error for the the TQL2 routine (see EISPACK).
            The no-error, normal completion code is zero.
        """
        super().__init__('''
            int eigen_decompose_real_symmetric_matrix(int n, double* A, double* w, double* Z, double* scratch){
                
                // tri-diagonalize the matrix
                eispack_tred2 ( n, A, w, scratch, Z );
                
                // QR decompose the tri-diagonal
                return eispack_tql2 ( n, w, scratch, Z );
            }
        ''', dependencies=[eispack_tred2(), eispack_tql2()])


class pseudo_inverse_real_symmetric_matrix_upper_triangular(SimpleCLLibrary):

    def __init__(self):
        """Compute the pseudo-inverse of a real symmetric matrix stored as an upper triangular vector.

        Results are placed in the upper triangular input vector.

        Parameters:
            n: the size of the symmetric matrix
            A[n*(n+1)/2]: on input, the matrix to inverse. On output, the inverse of the matrix.
            scratch[2*n + 2*n*n]: scratch data
        """
        super().__init__('''
            void pseudo_inverse_real_symmetric_matrix_upper_triangular(uint n, double* A, double* scratch){
                int i, j, z, ind_counter;
                double sum;
                
                double* w = scratch;
                double* Z = w + n;
                double* decomposition_scratch = Z + n * n;
                double* Z_transpose_w = decomposition_scratch + n;
                
                ind_counter = 0;
                for (j = 0; j < n; j++){
                    for (i = j; i < n; i++){
                        Z[i+j*n] = A[ind_counter];
                        ind_counter++;
                    }
                }

                eigen_decompose_real_symmetric_matrix(n, Z, w, Z, decomposition_scratch);
                
                // inverse the eigenvalues                
                for(i = 0; i < n; i++){
                    if(w[i] == 0.){
                        w[i] = 0;
                    }
                    else{
                        w[i] = fabs(1/w[i]);
                    }
                }
                
                // prepare the left matrix Z.T*w
                for(i = 0; i < n; i++){
                    for(j = 0; j < n; j++){
                        Z_transpose_w[i * n + j] = Z[j * n + i] * w[j];
                    }
                }
                
                // matrix multiplication of (Z.T*w) * Z
                ind_counter = 0;
                for(i = 0; i < n; i++){
                    for(j = i; j < n; j++){

                        sum = 0;
                        for(z = 0; z < n; z++){
                            sum += Z_transpose_w[i * n + z] * Z[z * n + j];
                        }

                        A[ind_counter] = sum;
                        ind_counter++;
                    }
                }
            }
        ''', dependencies=[eigen_decompose_real_symmetric_matrix()])


class Powell(SimpleCLLibraryFromFile):

    def __init__(self, eval_func, nmr_parameters, patience=2, patience_line_search=None,
                 reset_method='EXTRAPOLATED_POINT', **kwargs):
        """The Powell CL implementation.

        Args:
            eval_func (mot.lib.cl_function.CLFunction): the function we want to optimize, Should be of signature:
                ``double evaluate(local mot_float_type* x, void* data_void);``
            nmr_parameters (int): the number of parameters in the model, this will be hardcoded in the method
            patience (int): the patience of the Powell algorithm
            patience_line_search (int): the patience of the line search algorithm. If None, we set it equal to the
                patience.
            reset_method (str): one of ``RESET_TO_IDENTITY`` or ``EXTRAPOLATED_POINT``. The method used to
                reset the search directions every iteration.
        """
        dependencies = list(kwargs.get('dependencies', []))
        dependencies.append(eval_func)
        kwargs['dependencies'] = dependencies

        params = {
            'FUNCTION_NAME': eval_func.get_cl_function_name(),
            'NMR_PARAMS': nmr_parameters,
            'RESET_METHOD': reset_method.upper(),
            'PATIENCE': patience,
            'PATIENCE_LINE_SEARCH': patience if patience_line_search is None else patience_line_search
        }
        super().__init__(
            'int', 'powell', [
                'local mot_float_type* model_parameters',
                'void* data',
                'local mot_float_type* scratch_mot_float_type'
            ],
            resource_filename('mot', 'data/opencl/powell.cl'),
            var_replace_dict=params, **kwargs)

    def get_kernel_data(self):
        """Get the kernel data needed for this optimization routine to work."""
        return {
            'scratch_mot_float_type': LocalMemory(
                'mot_float_type',  3 * self._var_replace_dict['NMR_PARAMS'] + self._var_replace_dict['NMR_PARAMS']**2)
        }


class LibNMSimplex(SimpleCLLibraryFromFile):

    def __init__(self, function_name):
        """The NMSimplex algorithm as a reusable library component.

        Args:
            function_name (str): the name of the evaluation function to call, defaults to 'evaluate'.
                This should point to a function with signature:

                    ``double evaluate(local mot_float_type* x, void* data_void);``
        """
        params = {
            'FUNCTION_NAME': function_name
        }

        super().__init__(
            'int', 'lib_nmsimplex', [],
            resource_filename('mot', 'data/opencl/lib_nmsimplex.cl'),
            var_replace_dict=params)


class NMSimplex(SimpleCLLibrary):

    def __init__(self, function_name, nmr_parameters, patience=200, alpha=1.0, beta=0.5,
                 gamma=2.0, delta=0.5, scale=0.1, adaptive_scales=True, **kwargs):

        self._nmr_parameters = nmr_parameters

        if 'dependencies' in kwargs:
            kwargs['dependencies'] = list(kwargs['dependencies']) + [LibNMSimplex(function_name)]
        else:
            kwargs['dependencies'] = [LibNMSimplex(function_name)]

        params = {'NMR_PARAMS': nmr_parameters,
                  'PATIENCE': patience,
                  'ALPHA': alpha,
                  'BETA': beta,
                  'GAMMA': gamma,
                  'DELTA': delta,
                  'INITIAL_SIMPLEX_SCALES': '\n'.join('initial_simplex_scale[{}] = {};'.format(ind, scale)
                                                      for ind in range(nmr_parameters))}

        if adaptive_scales:
            params.update(
                {'ALPHA': 1,
                 'BETA': 0.75 - 1.0 / (2 * nmr_parameters),
                 'GAMMA': 1 + 2.0 / nmr_parameters,
                 'DELTA': 1 - 1.0 / nmr_parameters}
            )

        super().__init__('''
            int nmsimplex(local mot_float_type* model_parameters, void* data, 
                          local mot_float_type* initial_simplex_scale, 
                          local mot_float_type* nmsimplex_scratch){

                if(get_local_id(0) == 0){
                    %(INITIAL_SIMPLEX_SCALES)s
                }
                barrier(CLK_LOCAL_MEM_FENCE);

                mot_float_type fdiff;
                mot_float_type psi = 0;

                return lib_nmsimplex(%(NMR_PARAMS)r, model_parameters, data, initial_simplex_scale,
                                     &fdiff, psi, (int)(%(PATIENCE)r * (%(NMR_PARAMS)r+1)),
                                     %(ALPHA)r, %(BETA)r, %(GAMMA)r, %(DELTA)r,
                                     nmsimplex_scratch);
            }
        ''' % params, **kwargs)

    def get_kernel_data(self):
        """Get the kernel data needed for this optimization routine to work."""
        return {
            'nmsimplex_scratch': LocalMemory(
                'mot_float_type', self._nmr_parameters * 2 + (self._nmr_parameters + 1) ** 2 + 1),
            'initial_simplex_scale': LocalMemory('mot_float_type', self._nmr_parameters)
        }


class Subplex(SimpleCLLibraryFromFile):

    def __init__(self, eval_func, nmr_parameters, patience=10,
                 patience_nmsimplex=100, alpha=1.0, beta=0.5, gamma=2.0, delta=0.5, scale=1.0, psi=0.001, omega=0.01,
                 adaptive_scales=True, min_subspace_length='auto', max_subspace_length='auto', **kwargs):
        """The Subplex optimization routines."""
        dependencies = list(kwargs.get('dependencies', []))
        dependencies.append(eval_func)
        dependencies.append(LibNMSimplex('subspace_evaluate'))
        kwargs['dependencies'] = dependencies

        params = {
            'FUNCTION_NAME': eval_func.get_cl_function_name(),
            'PATIENCE': patience,
            'PATIENCE_NMSIMPLEX': patience_nmsimplex,
            'ALPHA': alpha,
            'BETA': beta,
            'GAMMA': gamma,
            'DELTA': delta,
            'PSI': psi,
            'OMEGA': omega,
            'NMR_PARAMS': nmr_parameters,
            'ADAPTIVE_SCALES': int(bool(adaptive_scales)),
            'MIN_SUBSPACE_LENGTH': (min(2, nmr_parameters) if min_subspace_length == 'auto' else min_subspace_length),
            'MAX_SUBSPACE_LENGTH': (min(5, nmr_parameters) if max_subspace_length == 'auto' else max_subspace_length)
        }

        s = ''
        for ind in range(nmr_parameters):
            s += 'initial_simplex_scale[{}] = {};'.format(ind, scale)
        params['INITIAL_SIMPLEX_SCALES'] = s

        super().__init__(
            'int', 'subplex', [
                'local mot_float_type* const model_parameters',
                'void* data',
                'local mot_float_type* initial_simplex_scale',
                'local mot_float_type* subplex_scratch_float',
                'local int* subplex_scratch_int'
            ],
            resource_filename('mot', 'data/opencl/subplex.cl'), var_replace_dict=params, **kwargs)

    def get_kernel_data(self):
        """Get the kernel data needed for this optimization routine to work."""
        return {
            'subplex_scratch_float': LocalMemory(
                'mot_float_type', 4 + self._var_replace_dict['NMR_PARAMS'] * 2
                                    + self._var_replace_dict['MAX_SUBSPACE_LENGTH'] * 2
                                    + (self._var_replace_dict['MAX_SUBSPACE_LENGTH'] * 2
                                       + self._var_replace_dict['MAX_SUBSPACE_LENGTH']+1)**2 + 1),
            'subplex_scratch_int': LocalMemory(
                'int', 2 + self._var_replace_dict['NMR_PARAMS']
                         + (self._var_replace_dict['NMR_PARAMS'] // self._var_replace_dict['MIN_SUBSPACE_LENGTH'])),
            'initial_simplex_scale': LocalMemory('mot_float_type', self._var_replace_dict['NMR_PARAMS'])
        }


class LevenbergMarquardt(SimpleCLLibraryFromFile):

    def __init__(self, eval_func, nmr_parameters, nmr_observations, jacobian_func, patience=250,
                 step_bound=100.0, scale_diag=1, usertol_mult=30, **kwargs):
        """The Powell CL implementation.

        Args:
            eval_func (mot.lib.cl_function.CLFunction): the function we want to optimize, Should be of signature:
                ``void evaluate(local mot_float_type* x, void* data_void, local mot_float_type* result);``
            nmr_parameters (int): the number of parameters in the model, this will be hardcoded in the method
            nmr_observations (int): the number of observations in the model
            jacobian_func (mot.lib.cl_function.CLFunction): the function used to compute the Jacobian.
            patience (int): the patience of the Powell algorithm
            patience_line_search (int): the patience of the line search algorithm
            reset_method (str): one of ``RESET_TO_IDENTITY`` or ``EXTRAPOLATED_POINT``. The method used to
                reset the search directions every iteration.
        """
        dependencies = list(kwargs.get('dependencies', []))
        dependencies.append(eval_func)
        dependencies.append(jacobian_func)
        kwargs['dependencies'] = dependencies

        var_replace_dict = {
            'FUNCTION_NAME': eval_func.get_cl_function_name(),
            'JACOBIAN_FUNCTION_NAME': jacobian_func.get_cl_function_name(),
            'NMR_PARAMS': nmr_parameters,
            'PATIENCE': patience,
            'NMR_OBSERVATIONS': nmr_observations,
            'SCALE_DIAG': int(bool(scale_diag)),
            'STEP_BOUND': step_bound,
            'USERTOL_MULT': usertol_mult
        }

        super().__init__(
            'int', 'lmmin', ['local mot_float_type* const model_parameters',
                             'void* data',
                             'mot_float_type* scratch_mot_float_type',
                             'int* scratch_int'],
            resource_filename('mot', 'data/opencl/lmmin.cl'),
            var_replace_dict=var_replace_dict, **kwargs)

    def get_kernel_data(self):
        """Get the kernel data needed for this optimization routine to work."""
        return {
            'scratch_mot_float_type': LocalMemory(
                'mot_float_type', 8 +
                                  2 * self._var_replace_dict['NMR_OBSERVATIONS'] +
                                  5 * self._var_replace_dict['NMR_PARAMS'] +
                                  self._var_replace_dict['NMR_PARAMS'] * self._var_replace_dict['NMR_OBSERVATIONS']),
            'scratch_int': LocalMemory('int', self._var_replace_dict['NMR_PARAMS'])
        }
